"""
Module with functions for creating plots of
color-magnitude  color-color diagrams.
"""

import json
import os
import pathlib
import warnings

from typing import Dict, List, Optional, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from matplotlib.colorbar import Colorbar
from matplotlib.ticker import MultipleLocator
from scipy.interpolate import interp1d
from typeguard import typechecked

from species.core.box import IsochroneBox, ColorMagBox, ColorColorBox
from species.read.read_filter import ReadFilter
from species.read.read_object import ReadObject
from species.util.core_util import print_section
from species.util.dust_util import calc_reddening, ism_extinction
from species.util.model_util import convert_model_name
from species.util.plot_util import (
    field_bounds_ticks,
    sptype_to_index,
    remove_color_duplicates,
)


@typechecked
def plot_color_magnitude(
    boxes: list,
    objects: Optional[
        Union[
            List[Tuple[str, str, str, str]],
            List[Tuple[str, str, str, str, Optional[dict], Optional[dict]]],
        ]
    ] = None,
    mass_labels: Optional[Dict[str, List[Tuple[float, str]]]] = None,
    teff_labels: Optional[Dict[str, List[Tuple[float, str]]]] = None,
    companion_labels: bool = False,
    accretion: bool = False,
    reddening: Optional[
        List[Tuple[Tuple[str, str], Tuple[str, float], str, float, Tuple[float, float]]]
    ] = None,
    ism_red: Optional[
        List[Tuple[Tuple[str, str], str, float, Tuple[float, float]]]
    ] = None,
    field_range: Optional[Tuple[str, str]] = None,
    label_x: str = "Color",
    label_y: str = "Absolute magnitude",
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    offset: Optional[Tuple[float, float]] = None,
    legend: Optional[Union[str, dict, Tuple[float, float]]] = "upper left",
    figsize: Optional[Tuple[float, float]] = (4.0, 4.8),
    output: Optional[str] = None,
) -> mpl.figure.Figure:
    """
    Function for creating a color-magnitude diagram.

    Parameters
    ----------
    boxes : list(species.core.box.ColorMagBox, species.core.box.IsochroneBox)
        Boxes with the color-magnitude and isochrone data from
        photometric libraries, spectral libraries, and/or atmospheric
        models. The synthetic data have to be created with
        :func:`~species.read.read_isochrone.ReadIsochrone.get_color_magnitude`.
        These boxes contain synthetic colors and magnitudes for a
        given age and a range of masses.
    objects : list(tuple(str, str, str, str)),
            list(tuple(str, str, str, str, dict, dict)), None
        Tuple with individual objects. The objects require a tuple with
        their database tag, the two filter names for the color, and the
        filter name for the absolute magnitude. Optionally, a
        dictionary with keyword arguments can be provided for the
        object's marker and label, respectively. For example,
        ``{'marker': 'o', 'ms': 10}`` for the marker and
        ``{'ha': 'left', 'va': 'bottom', 'xytext': (5, 5)})`` for the
        label. The parameter is not used if set to ``None``.
    mass_labels : dict(str, list(tuple(float, str))), None
        Plot labels with masses next to the isochrone data.
        The argument is a dictionary. The keys are the isochrone tags
        and the values are lists of tuples. Each tuple contains the
        mass in :math:`M_\\mathrm{J}` and the position of the label
        ('left' or 'right), for
        example ``{'sonora+0.5': [(10., 'left'), (20., 'right')]}``.
        No labels will be shown if the argument is set to ``None`` or
        if an isochrone tag is not included in the dictionary. The
        tags are stored as the ``iso_tag`` attribute of each
        :class:`~species.core.box.ColorColorBox`.
    teff_labels : dict(str, list(tuple(float, str))), None
        Plot labels with temperatures (K) next to the isochrones. The
        argument is a dictionary with the names of the models and a
        list with tuples that contain the effective temperatures and
        'left' or 'right' to indicate the position of the label (so
        similar to the use of mass_labels``). For example,
        ``{'planck': [(1000., 'left'), (1200., 'right')]``. No labels
        are shown if the argument is set to ``None``.
    companion_labels : bool
        Plot labels with the names of the directly imaged companions.
    accretion : bool
        Plot accreting, directly imaged objects with a different symbol
        than the regular, directly imaged objects. The object names
        from ``objects`` will be compared with the data from
        `data/companion_data/companion_data.json` to check if a
        companion is accreting or not.
    reddening : list(tuple(tuple(str, str), tuple(str, float),
            str, float, tuple(float, float))), None
        Include reddening arrows by providing a list with tuples. Each
        tuple contains the filter names for the color, the filter name
        and value of the magnitude, the mean particle radius (um), and
        the start position (color, mag) of the arrow in the plot, so
        ``((filter_color_1, filter_color_2), (filter_mag, mag_value),
        composition, radius, (x_pos, y_pos))``. The composition can be
        either ``'Fe'`` or ``'MgSiO3'`` (both with crystalline
        structure). A log-normal size distribution is used with the
        specified mean radius and the geometric standard deviation is
        fixed to 2. Both ``xlim`` and ``ylim`` need to be set for the
        correct rotation of the reddening label. The parameter is not
        used if set to ``None``.
    ism_red : list(tuple(tuple(str, str), str, float,
            tuple(float, float))), None
        List with reddening arrows for ISM extinction. Each item in the
        list is a tuple that itself contain a tuple with the filter
        names for the color, the filter name of the magnitude, the
        visual extinction, and the start position (color, mag) of the
        arrow in the plot, so ``((filter_color_1, filter_color_2),
        filter_mag, A_V, (x_pos, y_pos))``. The parameter is not used
        if the argument is set to ``None``.
    field_range : tuple(str, str), None
        Range of spectral types for which the field objects will be
        plotted and labeled at the discrete colorbar. The tuple
        should contain the lower and upper value of either the classes,
        for example ('K', 'T'), or as subclasses by including
        early/late, for example ('late K', 'early T'). The range is
        set to the default of 'early M' to 'early Y' if the
        argument is set to ``None``.
    label_x : str
        Label for the x-axis.
    label_y : str
        Label for the y-axis.
    xlim : tuple(float, float), None
        Limits for the x-axis. Not used if set to None.
    ylim : tuple(float, float), None
        Limits for the y-axis. Not used if set to None.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label.
    legend : str, tuple(float, float), dict, None
        Legend position or keyword arguments. No legend
        is shown if set to ``None``.
    figsize : tuple(float, float)
        Figure size.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for further
        customization of the plot.
    """

    print_section("Plot color-magnitude diagram")

    print("Boxes:")
    for item in boxes:
        if isinstance(item, list):
            item_type = item[0].__class__.__name__
            print(f"   - List with {len(item)} x {item_type}")
        else:
            print(f"   - {item.__class__.__name__}")

    if objects is None:
        print(f"\nObjects: {objects}")
    else:
        print("\nObjects:")
        for item in objects:
            print(f"   - {item[0]}: {item[1:]}")

    print(f"\nSpectral range: {field_range}")
    print(f"Companion labels: {companion_labels}")
    print(f"Accretion markers: {accretion}")

    print(f"\nMass labels: {mass_labels}")
    print(f"Teff labels: {teff_labels}")

    if reddening is None:
        print(f"\nReddening: {reddening}")
    else:
        print("\nReddening:")
        for item in reddening:
            print(f"   - {item}")

    print(f"\nFigure size: {figsize}")
    print(f"Limits x axis: {xlim}")
    print(f"Limits y axis: {ylim}")

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"

    # model_color = ("#234398", "#f6a432", "black")

    model_color = (
        "tab:blue",
        "tab:orange",
        "tab:green",
        "tab:red",
        "tab:purple",
        "tab:brown",
        "tab:pink",
        "tab:olive",
        "tab:cyan",
    )

    model_linestyle = ("-", "--", ":", "-.")

    if field_range is None:
        field_range = ("early M", "early Y")

    if (len(field_range[0]) == 1 and len(field_range[1]) != 1) or (
        len(field_range[0]) != 1 and len(field_range[1]) == 1
    ):
        raise ValueError(
            "The argument of 'field_range' should "
            "be a tuple with either the range of "
            "spectral type classes, for example "
            "('A', 'M'), or a tuple with the range "
            "of subclasses, for example ('early K', "
            "'late L')."
        )

    if len(field_range[0]) == 1 and len(field_range[1]) == 1:
        check_subclass = False
    else:
        check_subclass = True

    isochrones = []
    planck = []
    models = []
    empirical = []

    for item in boxes:
        if isinstance(item, IsochroneBox):
            isochrones.append(item)

        elif isinstance(item, ColorMagBox):
            if item.object_type == "model":
                models.append(item)

            elif item.library == "planck":
                planck.append(item)

            else:
                empirical.append(item)

        else:
            raise ValueError(
                f"Found a {type(item)} while only ColorMagBox and "
                "IsochroneBox objects can be provided to 'boxes'."
            )

    if empirical:
        fig = plt.figure(figsize=figsize)
        gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.5])
        gridsp.update(wspace=0.0, hspace=0.0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[2, 0])
        ax2 = plt.subplot(gridsp[0, 0])

    else:
        fig = plt.figure(figsize=figsize)
        gridsp = mpl.gridspec.GridSpec(1, 1)
        gridsp.update(wspace=0.0, hspace=0.0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[0, 0])

    ax1.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax1.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax1.xaxis.set_major_locator(MultipleLocator(1.0))
    ax1.yaxis.set_major_locator(MultipleLocator(1.0))

    ax1.xaxis.set_minor_locator(MultipleLocator(0.2))
    ax1.yaxis.set_minor_locator(MultipleLocator(0.2))

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    ax1.invert_yaxis()

    if offset is not None:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim is not None:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim is not None:
        ax1.set_ylim(ylim[0], ylim[1])

    if models is not None:
        count = 0

        model_dict = {}

        for j, item in enumerate(models):
            if item.library == "sonora-bobcat":
                model_key = item.library + item.iso_tag[-4:]
            else:
                model_key = item.library

            if model_key not in model_dict:
                model_dict[model_key] = [count, 0]
                count += 1

            else:
                model_dict[model_key] = [
                    model_dict[model_key][0],
                    model_dict[model_key][1] + 1,
                ]

            model_count = model_dict[model_key]

            if model_count[1] == 0:
                label = convert_model_name(item.library)

                if item.library == "sonora-bobcat":
                    metal = float(item.iso_tag[-4:])
                    label += f", [M/H] = {metal}"

                if item.library == "zhu2015":
                    ax1.plot(
                        item.color,
                        item.magnitude,
                        marker="x",
                        ms=5,
                        linestyle=model_linestyle[model_count[1]],
                        linewidth=0.6,
                        color="gray",
                        label=label,
                        zorder=0,
                    )

                    xlim = ax1.get_xlim()
                    ylim = ax1.get_ylim()

                    for i, teff_item in enumerate(item.sptype):
                        teff_label = (
                            rf"{teff_item:.0e} $M_\mathregular{{Jup}}^{2}$ yr$^{{-1}}$"
                        )

                        if item.magnitude[i] > ylim[1]:
                            ax1.annotate(
                                teff_label,
                                (item.color[i], item.magnitude[i]),
                                color="gray",
                                fontsize=8,
                                ha="left",
                                va="center",
                                xytext=(item.color[i] + 0.1, item.magnitude[i] + 0.05),
                                zorder=3,
                            )

                else:
                    ax1.plot(
                        item.color,
                        item.magnitude,
                        linestyle=model_linestyle[model_count[1]],
                        lw=1.0,
                        color=model_color[model_count[0]],
                        label=label,
                        zorder=0,
                    )

                    if mass_labels is not None:
                        interp_magnitude = interp1d(item.mass, item.magnitude)
                        interp_color = interp1d(item.mass, item.color)

                        if item.iso_tag in mass_labels:
                            label_select = mass_labels[item.iso_tag]
                        else:
                            label_select = []

                        for i, mass_item in enumerate(label_select):
                            if isinstance(mass_item, tuple):
                                mass_val = mass_item[0]
                                mass_pos = mass_item[1]

                            else:
                                mass_val = mass_item
                                mass_pos = "right"

                            pos_color = interp_color(mass_val)
                            pos_mag = interp_magnitude(mass_val)

                            # if j == 1 and mass_val == 10.:
                            #     mass_ha = "center"
                            #     mass_xytext = (pos_color, pos_mag-0.2)

                            if mass_pos == "left":
                                mass_ha = "right"
                                mass_xytext = (pos_color - 0.05, pos_mag)

                            else:
                                mass_ha = "left"
                                mass_xytext = (pos_color + 0.05, pos_mag)

                            mass_label = str(int(mass_val)) + r" M$_\mathregular{J}$"

                            xlim = ax1.get_xlim()
                            ylim = ax1.get_ylim()

                            if (
                                xlim[0] + 0.2 < pos_color < xlim[1] - 0.2
                                and ylim[1] + 0.2 < pos_mag < ylim[0] - 0.2
                            ):
                                ax1.scatter(
                                    pos_color,
                                    pos_mag,
                                    c=model_color[model_count[0]],
                                    s=15,
                                    edgecolor="none",
                                    zorder=0,
                                )

                                ax1.annotate(
                                    mass_label,
                                    (pos_color, pos_mag),
                                    color=model_color[model_count[0]],
                                    fontsize=9,
                                    xytext=mass_xytext,
                                    zorder=3,
                                    ha=mass_ha,
                                    va="center",
                                )

            else:
                ax1.plot(
                    item.color,
                    item.magnitude,
                    linestyle=model_linestyle[model_count[1]],
                    linewidth=0.6,
                    color=model_color[model_count[0]],
                    zorder=0,
                )

    if planck is not None:
        planck_count = 0

        for j, item in enumerate(planck):
            if planck_count == 0:
                label = convert_model_name(item.library)
            else:
                label = None

            ax1.plot(
                item.color,
                item.magnitude,
                linestyle="--",
                linewidth=0.8,
                color="gray",
                label=label,
                zorder=0,
            )

            if teff_labels is not None and planck_count == 0:
                interp_magnitude = interp1d(item.sptype, item.magnitude)
                interp_color = interp1d(item.sptype, item.color)

                if item.library in teff_labels:
                    label_select = teff_labels[item.library]
                else:
                    label_select = []

                for i, teff_item in enumerate(label_select):
                    if isinstance(teff_item, tuple):
                        teff_val = teff_item[0]
                        teff_pos = teff_item[1]

                    else:
                        teff_val = teff_item
                        teff_pos = "right"

                    if j == 0 or (j > 0 and teff_val < 20.0):
                        pos_color = interp_color(teff_val)
                        pos_mag = interp_magnitude(teff_val)

                        if teff_pos == "left":
                            teff_ha = "right"
                            teff_xytext = (pos_color - 0.05, pos_mag)

                        else:
                            teff_ha = "left"
                            teff_xytext = (pos_color + 0.05, pos_mag)

                        teff_label = f"{int(teff_val)} K"

                        xlim = ax1.get_xlim()
                        ylim = ax1.get_ylim()

                        if (
                            xlim[0] + 0.2 < pos_color < xlim[1] - 0.2
                            and ylim[1] + 0.2 < pos_mag < ylim[0] - 0.2
                        ):
                            ax1.scatter(
                                pos_color, pos_mag, c="gray", s=15, ec="none", zorder=0
                            )

                            if planck_count == 0:
                                ax1.annotate(
                                    teff_label,
                                    (pos_color, pos_mag),
                                    color="gray",
                                    fontsize=9,
                                    xytext=teff_xytext,
                                    zorder=3,
                                    ha=teff_ha,
                                    va="center",
                                )

            planck_count += 1

    if empirical:
        cmap = plt.cm.viridis

        bounds, ticks, ticklabels = field_bounds_ticks(field_range, check_subclass)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        for item in empirical:
            sptype = item.sptype
            color = item.color
            magnitude = item.magnitude
            names = item.names

            if isinstance(sptype, list):
                sptype = np.array(sptype)

            if item.object_type in ["field", None]:
                indices = np.where(sptype != "None")[0]

                sptype = sptype[indices]
                color = color[indices]
                magnitude = magnitude[indices]

                spt_disc = sptype_to_index(field_range, sptype, check_subclass)

                _, unique = np.unique(color, return_index=True)

                sptype = sptype[unique]
                color = color[unique]
                magnitude = magnitude[unique]
                spt_disc = spt_disc[unique]

                scat = ax1.scatter(
                    color,
                    magnitude,
                    c=spt_disc,
                    cmap=cmap,
                    norm=norm,
                    s=50,
                    alpha=0.7,
                    edgecolor="none",
                    zorder=2,
                )

                cb = Colorbar(
                    ax=ax2,
                    mappable=scat,
                    orientation="horizontal",
                    ticklocation="top",
                    format="%.2f",
                )

                cb.ax.tick_params(
                    width=1, length=5, labelsize=10, direction="in", color="black"
                )

                cb.set_ticks(ticks)
                cb.set_ticklabels(ticklabels)
                cb.ax.minorticks_off()

            elif item.object_type == "young":
                if objects is not None:
                    object_names = []

                    for obj_item in objects:
                        object_names.append(obj_item[0])

                    indices = remove_color_duplicates(object_names, names)

                    color = color[indices]
                    magnitude = magnitude[indices]

                ax1.plot(
                    color,
                    magnitude,
                    marker="s",
                    ms=4,
                    linestyle="none",
                    alpha=0.7,
                    color="gray",
                    markeredgecolor="black",
                    label="Young/low-gravity",
                    zorder=2,
                )

                # for item in names[indices]:
                #
                #     if item == '2MASSWJ2244316+204343':
                #         item = '2MASS 2244+2043'
                #
                #     kwargs = {'ha': 'left', 'va': 'center', 'fontsize': 8.5,
                #               'xytext': (5., 0.), 'color': 'black'}
                #
                #     ax1.annotate(item, (color, magnitude), zorder=3,
                #                  textcoords='offset points', **kwargs)

    if isochrones:
        for item in isochrones:
            ax1.plot(
                item.color, item.magnitude, linestyle="-", linewidth=1.0, color="black"
            )

    if reddening is not None:
        for item in reddening:
            ext_1, ext_2 = calc_reddening(
                item[0],
                item[1],
                composition=item[2],
                structure="crystalline",
                radius_g=item[3],
            )

            delta_x = ext_1 - ext_2
            delta_y = item[1][1]

            x_pos = item[4][0] + delta_x
            y_pos = item[4][1] + delta_y

            ax1.annotate(
                "",
                (x_pos, y_pos),
                xytext=(item[4][0], item[4][1]),
                fontsize=8,
                arrowprops={"arrowstyle": "->"},
                color="black",
                zorder=3.0,
            )

            x_pos_text = item[4][0] + delta_x / 2.0
            y_pos_text = item[4][1] + delta_y / 2.0

            vector_len = np.sqrt(delta_x**2 + delta_y**2)

            if item[2] == "MgSiO3":
                dust_species = r"MgSiO$_{3}$"
            elif item[2] == "Fe":
                dust_species = "Fe"

            if (item[3]).is_integer():
                red_label = f"{dust_species} ({item[3]:.0f} µm)"
            else:
                red_label = f"{dust_species} ({item[3]:.1f} µm)"

            text = ax1.annotate(
                red_label,
                (x_pos_text, y_pos_text),
                xytext=(7.0 * delta_y / vector_len, 7.0 * delta_x / vector_len),
                textcoords="offset points",
                fontsize=8.0,
                color="black",
                ha="center",
                va="center",
            )

            ax1.plot([item[4][0], x_pos], [item[4][1], y_pos], "-", color="white")

            sp1 = ax1.transData.transform_point((item[4][0], item[4][1]))
            sp2 = ax1.transData.transform_point((x_pos, y_pos))

            angle = np.degrees(np.arctan2(sp2[1] - sp1[1], sp2[0] - sp1[0]))
            text.set_rotation(angle)

    if ism_red is not None:
        for item in ism_red:
            # Color filters
            read_filt_0 = ReadFilter(item[0][0])
            read_filt_1 = ReadFilter(item[0][1])

            # Magnitude filter
            read_filt_2 = ReadFilter(item[1])

            mean_wavel = np.array(
                [
                    read_filt_0.mean_wavelength(),
                    read_filt_1.mean_wavelength(),
                    read_filt_2.mean_wavelength(),
                ]
            )

            ext_mag = ism_extinction(item[2], 3.1, mean_wavel)

            delta_x = ext_mag[0] - ext_mag[1]
            delta_y = ext_mag[2]

            x_pos = item[3][0] + delta_x
            y_pos = item[3][1] + delta_y

            ax1.annotate(
                "",
                (x_pos, y_pos),
                xytext=(item[3][0], item[3][1]),
                fontsize=8,
                arrowprops={"arrowstyle": "->"},
                color="black",
                zorder=3.0,
            )

            x_pos_text = item[3][0] + delta_x / 2.0
            y_pos_text = item[3][1] + delta_y / 2.0

            vector_len = np.sqrt(delta_x**2 + delta_y**2)

            if (item[2]).is_integer():
                red_label = rf"A$_\mathregular{{V}}$ = {item[2]:.0f}"
            else:
                red_label = rf"A$_\mathregular{{V}}$ = {item[2]:.1f}"

            text = ax1.annotate(
                red_label,
                (x_pos_text, y_pos_text),
                xytext=(8.0 * delta_y / vector_len, 8.0 * delta_x / vector_len),
                textcoords="offset points",
                fontsize=8.0,
                color="black",
                ha="center",
                va="center",
            )

            ax1.plot([item[3][0], x_pos], [item[3][1], y_pos], "-", color="white")

            sp1 = ax1.transData.transform_point((item[3][0], item[3][1]))
            sp2 = ax1.transData.transform_point((x_pos, y_pos))

            angle = np.degrees(np.arctan2(sp2[1] - sp1[1], sp2[0] - sp1[0]))
            text.set_rotation(angle)

    if objects is not None:
        for i, item in enumerate(objects):
            objdata = ReadObject(item[0])

            objcolor1 = objdata.get_photometry(item[1])
            objcolor2 = objdata.get_photometry(item[2])

            if objcolor1.ndim == 2:
                print(
                    f"Found {objcolor1.shape[1]} values for filter {item[1]} of {item[0]}"
                )
                print(
                    f"so using the first value:  {objcolor1[0, 0]} +/- {objcolor1[1, 0]} mag"
                )
                objcolor1 = objcolor1[:, 0]

            if objcolor2.ndim == 2:
                print(
                    f"Found {objcolor2.shape[1]} values for filter {item[2]} of {item[0]}"
                )
                print(
                    f"so using the first value:  {objcolor2[0, 0]} +/- {objcolor2[1, 0]} mag"
                )
                objcolor2 = objcolor2[:, 0]

            abs_mag, abs_err = objdata.get_absmag(item[3])

            if isinstance(abs_mag, np.ndarray):
                abs_mag = abs_mag[0]
                abs_err = abs_err[0]

            colorerr = np.sqrt(objcolor1[1] ** 2 + objcolor2[1] ** 2)
            x_color = objcolor1[0] - objcolor2[0]

            species_folder = str(pathlib.Path(__file__).parent.resolve())[:-4]
            data_file = os.path.join(
                species_folder, "data/companion_data/companion_data.json"
            )

            with open(data_file, "r", encoding="utf-8") as json_file:
                comp_data = json.load(json_file)

            if len(item) > 4 and item[4] is not None:
                kwargs = item[4]

            else:
                kwargs = {
                    "marker": "o",
                    "ms": 5.0,
                    "color": "black",
                    "mfc": "white",
                    "mec": "black",
                    "label": "Companions",
                }

                if (
                    accretion
                    and item[0] in comp_data
                    and comp_data[item[0]]["accretion"]
                ):
                    kwargs["marker"] = "X"
                    kwargs["ms"] = 7.0
                    kwargs["label"] = "Accreting"

            ax1.errorbar(
                x_color, abs_mag, yerr=abs_err, xerr=colorerr, zorder=3, **kwargs
            )

            if companion_labels:
                if len(item) > 4 and item[5] is not None:
                    kwargs = item[5]

                else:
                    kwargs = {
                        "ha": "left",
                        "va": "bottom",
                        "fontsize": 8.5,
                        "xytext": (5.0, 5.0),
                        "color": "black",
                    }

                ax1.annotate(
                    objdata.object_name,
                    (x_color, abs_mag),
                    zorder=3,
                    textcoords="offset points",
                    **kwargs,
                )

    if legend is not None:
        handles, labels = ax1.get_legend_handles_labels()

        # Prevent duplicates
        by_label = dict(zip(labels, handles))

        if handles:
            if isinstance(legend, (str, tuple)):
                ax1.legend(
                    by_label.values(),
                    by_label.keys(),
                    loc=legend,
                    fontsize=8.5,
                    frameon=False,
                    numpoints=1,
                )

            else:
                ax1.legend(by_label.values(), by_label.keys(), **legend)

    if output is None:
        plt.show()
    else:
        print(f"\nOutput: {output}")
        plt.savefig(output, bbox_inches="tight")

    return fig


@typechecked
def plot_color_color(
    boxes: list,
    objects: Optional[
        Union[
            List[Tuple[str, Tuple[str, str], Tuple[str, str]]],
            List[
                Tuple[
                    str,
                    Tuple[str, str],
                    Tuple[str, str],
                    Optional[dict],
                    Optional[dict],
                ]
            ],
        ]
    ] = None,
    mass_labels: Optional[Dict[str, List[Tuple[float, str]]]] = None,
    teff_labels: Optional[Dict[str, List[Tuple[float, str]]]] = None,
    companion_labels: bool = False,
    reddening: Optional[
        List[
            Tuple[
                Tuple[str, str],
                Tuple[str, str],
                Tuple[str, float],
                str,
                float,
                Tuple[float, float],
            ]
        ]
    ] = None,
    field_range: Optional[Tuple[str, str]] = None,
    label_x: str = "Color",
    label_y: str = "Color",
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    offset: Optional[Tuple[float, float]] = None,
    legend: Optional[Union[str, dict, Tuple[float, float]]] = "upper left",
    figsize: Optional[Tuple[float, float]] = (4.0, 4.3),
    output: Optional[str] = None,
) -> mpl.figure.Figure:
    """
    Function for creating a color-color diagram.

    Parameters
    ----------
    boxes : list(species.core.box.ColorColorBox, species.core.box.IsochroneBox)
        Boxes with the color-color from photometric libraries,
        spectral libraries, isochrones, and/or atmospheric models.
    objects : tuple(tuple(str, tuple(str, str), tuple(str, str))),
            tuple(tuple(str, tuple(str, str), tuple(str, str), dict, dict)), None
        Tuple with individual objects. The objects require a tuple
        with their database tag, the two filter names for the first
        color, and the two filter names for the second color.
        Optionally, a dictionary with keyword arguments can be provided
        for the object's marker and label, respectively. For
        example, ``{'marker': 'o', 'ms': 10}`` for the marker
        and ``{'ha': 'left', 'va': 'bottom', 'xytext': (5, 5)})``
        for the label. The parameter is not used if set to ``None``.
    mass_labels : dict(str, list(tuple(float, str))), None
        Plot labels with masses next to the isochrone data.
        The argument is a dictionary. The keys are the isochrone tags
        and the values are lists of tuples. Each tuple contains the
        mass in :math:`M_\\mathrm{J}` and the position of the label
        ('left' or 'right), for
        example ``{'sonora+0.5': [(10., 'left'), (20., 'right')]}``.
        No labels will be shown if the argument is set to ``None`` or
        if an isochrone tag is not included in the dictionary. The
        tags are stored as the ``iso_tag`` attribute of each
        :class:`~species.core.box.ColorColorBox`.
    teff_labels : dict(str, list(tuple(float, str))), None
        Plot labels with temperatures (K) next to the isochrones. The
        argument is a dictionary with the names of the models and a
        list with tuples that contain the effective temperatures and
        'left' or 'right' to indicate the position of the label (so
        similar to the use of mass_labels``). For example,
        ``{'planck': [(1000., 'left'), (1200., 'right')]``. No labels
        are shown if the argument is set to ``None``.
    companion_labels : bool
        Plot labels with the names of the directly imaged companions.
    reddening : list(tuple(tuple(str, str), tuple(str, str),
            tuple(str, float), str, float, tuple(float, float)), None
        Include reddening arrows by providing a list with tuples.
        Each tuple contains the filter names for the color, the filter
        name for the magnitude, the particle radius (um), and the start
        position (color, mag) of the arrow in the plot, so
        (filter_color_1, filter_color_2, filter_mag, composition,
        radius, (x_pos, y_pos)). The composition can be either 'Fe' or
        'MgSiO3' (both with crystalline structure). The parameter is
        not used if set to ``None``.
    field_range : tuple(str, str), None
        Range of spectral types for which the field objects will be
        plotted and labeled at the discrete colorbar. The tuple
        should contain the lower and upper value of either the classes,
        for example ('K', 'T'), or as subclasses by including
        early/late, for example ('late K', 'early T'). The range is
        set to the default of 'early M' to 'early Y' if the
        argument is set to ``None``.
    label_x : str
        Label for the x-axis.
    label_y : str
        Label for the y-axis.
    xlim : tuple(float, float)
        Limits for the x-axis.
    ylim : tuple(float, float)
        Limits for the y-axis.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label.
    legend : str, tuple(float, float), dict, None
        Legend position or dictionary with keyword arguments.
        No legend is shown if the argument is set to ``None``.
    figsize : tuple(float, float)
        Figure size.
    output : str, None
        Output filename for the plot. The plot is shown in an
        interface window if the argument is set to ``None``.

    Returns
    -------
    matplotlib.figure.Figure
        The ``Figure`` object that can be used for further
        customization of the plot.
    """

    print_section("Plot color-color diagram")

    print("Boxes:")
    for item in boxes:
        if isinstance(item, list):
            item_type = item[0].__class__.__name__
            print(f"   - List with {len(item)} x {item_type}")
        else:
            print(f"   - {item.__class__.__name__}")

    if objects is None:
        print(f"\nObjects: {objects}")
    else:
        print("\nObjects:")
        for item in objects:
            print(f"   - {item[0]}: {item[1:]}")

    print(f"\nSpectral range: {field_range}")
    print(f"Companion labels: {companion_labels}")

    print(f"\nMass labels: {mass_labels}")
    print(f"Teff labels: {teff_labels}")

    if reddening is None:
        print(f"\nReddening: {reddening}")
    else:
        print("\nReddening:")
        for item in reddening:
            print(f"   - {item}")

    print(f"\nFigure size: {figsize}")
    print(f"Limits x axis: {xlim}")
    print(f"Limits y axis: {ylim}")

    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "dejavuserif"

    # model_color = ("#234398", "#f6a432", "black")

    model_color = (
        "tab:blue",
        "tab:orange",
        "tab:green",
        "tab:red",
        "tab:purple",
        "tab:brown",
        "tab:pink",
        "tab:olive",
        "tab:cyan",
    )

    model_linestyle = ("-", "--", ":", "-.")

    if field_range is None:
        field_range = ("early M", "early Y")

    if (len(field_range[0]) == 1 and len(field_range[1]) != 1) or (
        len(field_range[0]) != 1 and len(field_range[1]) == 1
    ):
        raise ValueError(
            "The argument of 'field_range' should "
            "be a tuple with either the range of "
            "spectral type classes, for example "
            "('A', 'M'), or a tuple with the range "
            "of subclasses, for example ('early K', "
            "'late L')."
        )

    if len(field_range[0]) == 1 and len(field_range[1]) == 1:
        check_subclass = False
    else:
        check_subclass = True

    isochrones = []
    planck = []
    models = []
    empirical = []
    spectra = []

    for item in boxes:
        if isinstance(item, IsochroneBox):
            isochrones.append(item)

        elif isinstance(item, ColorColorBox):
            if item.object_type == "model":
                models.append(item)

            elif item.object_type == "spectra":
                spectra.append(item)

            elif item.library == "planck":
                planck.append(item)

            else:
                empirical.append(item)

        else:
            raise ValueError(
                f"Found a {type(item)} while only ColorColorBox and "
                "IsochroneBox objects can be provided to 'boxes'."
            )

    fig = plt.figure(figsize=figsize)

    if empirical:
        gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.0])
    else:
        gridsp = mpl.gridspec.GridSpec(1, 1)

    gridsp.update(wspace=0.0, hspace=0.0, left=0, right=1, bottom=0, top=1)

    if empirical:
        ax1 = plt.subplot(gridsp[2, 0])
        ax2 = plt.subplot(gridsp[0, 0])
    else:
        ax1 = plt.subplot(gridsp[0, 0])
        ax2 = None

    ax1.tick_params(
        axis="both",
        which="major",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=5,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax1.tick_params(
        axis="both",
        which="minor",
        colors="black",
        labelcolor="black",
        direction="in",
        width=1,
        length=3,
        labelsize=12,
        top=True,
        bottom=True,
        left=True,
        right=True,
    )

    ax1.xaxis.set_major_locator(MultipleLocator(0.5))
    ax1.yaxis.set_major_locator(MultipleLocator(0.5))

    ax1.xaxis.set_minor_locator(MultipleLocator(0.1))
    ax1.yaxis.set_minor_locator(MultipleLocator(0.1))

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    if offset:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim:
        ax1.set_ylim(ylim[0], ylim[1])

    if models is not None:
        count = 0

        model_dict = {}

        for j, item in enumerate(models):
            if item.library == "sonora-bobcat":
                model_key = item.library + item.iso_tag[-4:]
            else:
                model_key = item.library

            if model_key not in model_dict:
                model_dict[model_key] = [count, 0]
                count += 1

            else:
                model_dict[model_key] = [
                    model_dict[model_key][0],
                    model_dict[model_key][1] + 1,
                ]

            model_count = model_dict[model_key]

            if model_count[1] == 0:
                label = convert_model_name(item.library)

                if item.library == "sonora-bobcat":
                    metal = float(item.iso_tag[-4:])
                    label += f", [M/H] = {metal}"

                if item.library == "zhu2015":
                    ax1.plot(
                        item.color1,
                        item.color2,
                        marker="x",
                        ms=5,
                        linestyle=model_linestyle[model_count[1]],
                        linewidth=0.6,
                        color="gray",
                        label=label,
                        zorder=0,
                    )

                    xlim = ax1.get_xlim()
                    ylim = ax1.get_ylim()

                    for i, teff_item in enumerate(item.sptype):
                        teff_label = (
                            rf"{teff_item:.0e} $M_\mathregular{{Jup}}^{2}$ yr$^{{-1}}$"
                        )

                        if item.color2[i] < ylim[1]:
                            ax1.annotate(
                                teff_label,
                                (item.color1[i], item.color2[i]),
                                color="gray",
                                fontsize=8,
                                ha="left",
                                va="center",
                                xytext=(item.color1[i] + 0.1, item.color2[i] - 0.05),
                                zorder=3,
                            )

                else:
                    ax1.plot(
                        item.color1,
                        item.color2,
                        linestyle=model_linestyle[model_count[1]],
                        lw=1.0,
                        color=model_color[model_count[0]],
                        label=label,
                        zorder=0,
                    )

                    if mass_labels is not None:
                        interp_color1 = interp1d(item.mass, item.color1)
                        interp_color2 = interp1d(item.mass, item.color2)

                        if item.iso_tag in mass_labels:
                            label_select = mass_labels[item.iso_tag]
                        else:
                            label_select = []

                        for i, mass_item in enumerate(label_select):
                            mass_val = mass_item[0]
                            mass_pos = mass_item[1]

                            pos_color1 = interp_color1(mass_val)
                            pos_color2 = interp_color2(mass_val)

                            if mass_pos == "left":
                                mass_ha = "right"
                                mass_xytext = (pos_color1 - 0.05, pos_color2)

                            else:
                                mass_ha = "left"
                                mass_xytext = (pos_color1 + 0.05, pos_color2)

                            mass_label = str(int(mass_val)) + r" M$_\mathregular{J}$"

                            xlim = ax1.get_xlim()
                            ylim = ax1.get_ylim()

                            if (
                                xlim[0] + 0.2 < pos_color1 < xlim[1] - 0.2
                                and ylim[0] + 0.2 < pos_color2 < ylim[1] - 0.2
                            ):
                                ax1.scatter(
                                    pos_color1,
                                    pos_color2,
                                    c=model_color[model_count[0]],
                                    s=15,
                                    edgecolor="none",
                                    zorder=0,
                                )

                                ax1.annotate(
                                    mass_label,
                                    (pos_color1, pos_color2),
                                    color=model_color[model_count[0]],
                                    fontsize=9,
                                    xytext=mass_xytext,
                                    ha=mass_ha,
                                    va="center",
                                    zorder=3,
                                )

                            else:
                                warnings.warn(
                                    f"Please use larger axes limits "
                                    f"to include the mass label for "
                                    f"{mass_val} Mjup."
                                )

            else:
                ax1.plot(
                    item.color1,
                    item.color2,
                    linestyle=model_linestyle[model_count[1]],
                    linewidth=0.6,
                    color=model_color[model_count[0]],
                    label=None,
                    zorder=0,
                )

    if planck is not None:
        planck_count = 0

        for j, item in enumerate(planck):
            if planck_count == 0:
                label = convert_model_name(item.library)

                ax1.plot(
                    item.color1,
                    item.color2,
                    ls="--",
                    linewidth=0.8,
                    color="gray",
                    label=label,
                    zorder=0,
                )

                if teff_labels is not None:
                    interp_color1 = interp1d(item.sptype, item.color1)
                    interp_color2 = interp1d(item.sptype, item.color2)

                    if item.library in teff_labels:
                        label_select = teff_labels[item.library]
                    else:
                        label_select = []

                    for i, teff_item in enumerate(label_select):
                        if isinstance(teff_item, tuple):
                            teff_val = teff_item[0]
                            teff_pos = teff_item[1]

                        else:
                            teff_val = teff_item
                            teff_pos = "right"

                        if j == 0 or (j > 0 and teff_val < 20.0):
                            pos_color1 = interp_color1(teff_val)
                            pos_color2 = interp_color2(teff_val)

                            if teff_pos == "left":
                                teff_ha = "right"
                                teff_xytext = (pos_color1 - 0.05, pos_color2)

                            else:
                                teff_ha = "left"
                                teff_xytext = (pos_color1 + 0.05, pos_color2)

                            teff_label = f"{int(teff_val)} K"

                            xlim = ax1.get_xlim()
                            ylim = ax1.get_ylim()

                            if (
                                xlim[0] + 0.2 < pos_color1 < xlim[1] - 0.2
                                and ylim[0] + 0.2 < pos_color2 < ylim[1] - 0.2
                            ):
                                ax1.scatter(
                                    pos_color1,
                                    pos_color2,
                                    c="gray",
                                    s=15,
                                    edgecolor="none",
                                    zorder=0,
                                )

                                ax1.annotate(
                                    teff_label,
                                    (pos_color1, pos_color2),
                                    color="gray",
                                    fontsize=9,
                                    xytext=teff_xytext,
                                    zorder=3,
                                    ha=teff_ha,
                                    va="center",
                                )

            else:
                ax1.plot(
                    item.color1, item.color2, ls="--", lw=0.5, color="gray", zorder=0
                )

            planck_count += 1

    if spectra is not None:
        spectra_count = 0

        for j, item in enumerate(spectra):
            if spectra_count == 0:
                label = convert_model_name(item.library)

                ax1.plot(
                    item.color1,
                    item.color2,
                    ls="--",
                    linewidth=0.8,
                    color="tomato",
                    label=label,
                    zorder=0,
                )

                if teff_labels is not None:
                    interp_color1 = interp1d(item.sptype, item.color1)
                    interp_color2 = interp1d(item.sptype, item.color2)

                    if item.library in teff_labels:
                        label_select = teff_labels[item.library]
                    else:
                        label_select = []

                    for i, teff_item in enumerate(label_select):
                        if isinstance(teff_item, tuple):
                            teff_val = teff_item[0]
                            teff_pos = teff_item[1]

                        else:
                            teff_val = teff_item
                            teff_pos = "right"

                        if j == 0 or (j > 0 and teff_val < 20.0):
                            pos_color1 = interp_color1(teff_val)
                            pos_color2 = interp_color2(teff_val)

                            if teff_pos == "left":
                                teff_ha = "right"
                                teff_xytext = (pos_color1 - 0.05, pos_color2)

                            else:
                                teff_ha = "left"
                                teff_xytext = (pos_color1 + 0.05, pos_color2)

                            teff_label = f"{int(teff_val)} K"

                            xlim = ax1.get_xlim()
                            ylim = ax1.get_ylim()

                            if (
                                xlim[0] + 0.2 < pos_color1 < xlim[1] - 0.2
                                and ylim[0] + 0.2 < pos_color2 < ylim[1] - 0.2
                            ):
                                ax1.scatter(
                                    pos_color1,
                                    pos_color2,
                                    c="tomato",
                                    s=15,
                                    edgecolor="none",
                                    zorder=0,
                                )

                                ax1.annotate(
                                    teff_label,
                                    (pos_color1, pos_color2),
                                    color="tomato",
                                    fontsize=9,
                                    xytext=teff_xytext,
                                    zorder=3,
                                    ha=teff_ha,
                                    va="center",
                                )

            else:
                ax1.plot(
                    item.color1, item.color2, ls="--", lw=0.5, color="tomato", zorder=0
                )

            spectra_count += 1

    if empirical:
        cmap = plt.cm.viridis

        bounds, ticks, ticklabels = field_bounds_ticks(field_range, check_subclass)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        for item in empirical:
            sptype = item.sptype
            names = item.names
            color1 = item.color1
            color2 = item.color2

            if isinstance(sptype, list):
                sptype = np.array(sptype)

            if item.object_type in ["field", None]:
                indices = np.where(sptype != "None")[0]

                sptype = sptype[indices]
                color1 = color1[indices]
                color2 = color2[indices]

                spt_disc = sptype_to_index(field_range, sptype, check_subclass)
                _, unique = np.unique(color1, return_index=True)

                sptype = sptype[unique]
                color1 = color1[unique]
                color2 = color2[unique]
                spt_disc = spt_disc[unique]

                scat = ax1.scatter(
                    color1,
                    color2,
                    c=spt_disc,
                    cmap=cmap,
                    norm=norm,
                    s=50,
                    alpha=0.7,
                    edgecolor="none",
                    zorder=2,
                )

                cb = Colorbar(
                    ax=ax2,
                    mappable=scat,
                    orientation="horizontal",
                    ticklocation="top",
                    format="%.2f",
                )

                cb.ax.tick_params(
                    width=1, length=5, labelsize=10, direction="in", color="black"
                )

                cb.set_ticks(ticks)
                cb.set_ticklabels(ticklabels)
                cb.ax.minorticks_off()

            elif item.object_type == "young":
                if objects is not None:
                    object_names = []

                    for obj_item in objects:
                        object_names.append(obj_item[0])

                    indices = remove_color_duplicates(object_names, names)

                    color1 = color1[indices]
                    color2 = color2[indices]

                ax1.plot(
                    color1,
                    color2,
                    marker="s",
                    ms=4,
                    linestyle="none",
                    alpha=0.7,
                    color="gray",
                    markeredgecolor="black",
                    label="Young/low-gravity",
                    zorder=2,
                )

    if isochrones:
        for item in isochrones:
            ax1.plot(
                item.colors[0],
                item.colors[1],
                linestyle="-",
                linewidth=1.0,
                color="black",
            )

    if reddening is not None:
        for item in reddening:
            ext_1, ext_2 = calc_reddening(
                item[0],
                item[2],
                composition=item[3],
                structure="crystalline",
                radius_g=item[4],
            )

            ext_3, ext_4 = calc_reddening(
                item[1],
                item[2],
                composition=item[3],
                structure="crystalline",
                radius_g=item[4],
            )

            delta_x = ext_1 - ext_2
            delta_y = ext_3 - ext_4

            x_pos = item[5][0] + delta_x
            y_pos = item[5][1] + delta_y

            ax1.annotate(
                "",
                (x_pos, y_pos),
                xytext=(item[5][0], item[5][1]),
                fontsize=8,
                arrowprops={"arrowstyle": "->"},
                color="black",
                zorder=3.0,
            )

            x_pos_text = item[5][0] + delta_x / 2.0
            y_pos_text = item[5][1] + delta_y / 2.0

            vector_len = np.sqrt(delta_x**2 + delta_y**2)

            if item[3] == "MgSiO3":
                dust_species = r"MgSiO$_{3}$"

            elif item[3] == "Fe":
                dust_species = "Fe"

            if item[4].is_integer():
                red_label = f"{dust_species} ({item[4]:.0f} µm)"

            else:
                red_label = f"{dust_species} ({item[4]:.1f} µm)"

            text = ax1.annotate(
                red_label,
                (x_pos_text, y_pos_text),
                xytext=(-7.0 * delta_y / vector_len, 7.0 * delta_x / vector_len),
                textcoords="offset points",
                fontsize=8.0,
                color="black",
                ha="center",
                va="center",
            )

            ax1.plot([item[5][0], x_pos], [item[5][1], y_pos], "-", color="white")

            sp1 = ax1.transData.transform_point((item[5][0], item[5][1]))
            sp2 = ax1.transData.transform_point((x_pos, y_pos))

            angle = np.degrees(np.arctan2(sp2[1] - sp1[1], sp2[0] - sp1[0]))
            text.set_rotation(angle)

    if objects is not None:
        for i, item in enumerate(objects):
            objdata = ReadObject(item[0])

            objphot1 = objdata.get_photometry(item[1][0])
            objphot2 = objdata.get_photometry(item[1][1])
            objphot3 = objdata.get_photometry(item[2][0])
            objphot4 = objdata.get_photometry(item[2][1])

            if objphot1.ndim == 2:
                print(
                    f"Found {objphot1.shape[1]} values for "
                    f"filter {item[1][0]} of {item[0]} "
                    f"so using the first magnitude: "
                    f"{objphot1[0, 0]} +/- {objphot1[1, 0]}"
                )

                objphot1 = objphot1[:, 0]

            if objphot2.ndim == 2:
                print(
                    f"Found {objphot2.shape[1]} values for "
                    f"filter {item[1][1]} of {item[0]} "
                    f"so using the first magnitude: "
                    f"{objphot2[0, 0]} +/- {objphot2[1, 0]}"
                )

                objphot2 = objphot2[:, 0]

            if objphot3.ndim == 2:
                print(
                    f"Found {objphot3.shape[1]} values for "
                    f"filter {item[2][0]} of {item[0]} "
                    f"so using the first magnitude: "
                    f"{objphot3[0, 0]} +/- {objphot3[1, 0]}"
                )

                objphot3 = objphot3[:, 0]

            if objphot4.ndim == 2:
                print(
                    f"Found {objphot4.shape[1]} values for "
                    f"filter {item[2][1]} of {item[0]} "
                    f"so using the first magnitude: "
                    f"{objphot4[0, 0]} +/- {objphot4[1, 0]}"
                )

                objphot4 = objphot4[:, 0]

            color1 = objphot1[0] - objphot2[0]
            color2 = objphot3[0] - objphot4[0]

            error1 = np.sqrt(objphot1[1] ** 2 + objphot2[1] ** 2)
            error2 = np.sqrt(objphot3[1] ** 2 + objphot4[1] ** 2)

            if len(item) > 3 and item[3] is not None:
                kwargs = item[3]

            else:
                kwargs = {
                    "marker": "o",
                    "ms": 5.0,
                    "color": "black",
                    "mfc": "white",
                    "mec": "black",
                    "label": "Companions",
                }

            ax1.errorbar(color1, color2, xerr=error1, yerr=error2, zorder=3, **kwargs)

            if companion_labels:
                if len(item) > 3 and item[4] is not None:
                    kwargs = item[4]

                else:
                    kwargs = {
                        "ha": "left",
                        "va": "bottom",
                        "fontsize": 8.5,
                        "xytext": (5.0, 5.0),
                        "color": "black",
                    }

                ax1.annotate(
                    objdata.object_name,
                    (color1, color2),
                    zorder=3,
                    textcoords="offset points",
                    **kwargs,
                )

    handles, labels = ax1.get_legend_handles_labels()

    if legend is not None:
        handles, labels = ax1.get_legend_handles_labels()

        # Prevent duplicates
        by_label = dict(zip(labels, handles))

        if handles:
            if isinstance(legend, (str, tuple)):
                ax1.legend(
                    by_label.values(),
                    by_label.keys(),
                    loc=legend,
                    fontsize=8.5,
                    frameon=False,
                    numpoints=1,
                )

            else:
                ax1.legend(by_label.values(), by_label.keys(), **legend)

    if output is None:
        plt.show()
    else:
        print(f"\nOutput: {output}")
        plt.savefig(output, bbox_inches="tight")

    return fig



import os
import urllib.request

import numpy as np

from species.core import constants
from species.util.data_util import extract_tarfile


def add_sonora(database, input_path):
    """
    Function for adding the
    `Sonora Bobcat <https://zenodo.org/record/5063476>`_
    isochrone data to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    url = "https://zenodo.org/record/5063476/files/evolution_and_photometery.tar.gz"

    input_file = "evolution_and_photometery.tar.gz"
    data_file = os.path.join(input_path, input_file)
    sub_folder = input_file.split(".", maxsplit=1)[0]
    data_folder = os.path.join(input_path, sub_folder)

    if not os.path.exists(data_folder):
        os.makedirs(data_folder)

    if not os.path.isfile(data_file):
        print("Downloading Sonora Bobcat evolution (929 kB)...", end="", flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(" [DONE]")

    print("Unpacking Sonora Bobcat evolution (929 kB)...", end="", flush=True)
    extract_tarfile(data_file, data_folder)
    print(" [DONE]")

    iso_files = [
        "evo_tables+0.0/nc+0.0_co1.0_age",
        "evo_tables+0.5/nc+0.5_co1.0_age",
        "evo_tables-0.5/nc-0.5_co1.0_age",
    ]

    labels = ["[M/H] = +0.0", "[M/H] = +0.5", "[M/H] = -0.5"]

    for i, item in enumerate(iso_files):
        iso_file = f"evolution_tables/{item}"
        iso_path = os.path.join(data_folder, iso_file)

        iso_data = []

        with open(iso_path, encoding="utf-8") as open_file:
            for j, line in enumerate(open_file):
                if j == 0 or " " not in line.strip():
                    continue

                # age(Gyr)  M/Msun  log(L/Lsun)  Teff(K)  log(g)  R/Rsun
                param = list(filter(None, line.strip().split(" ")))
                param = list(map(float, param))

                param[0] = 1e3 * param[0]  # (Gyr) -> (Myr)
                param[1] = (
                    param[1] * constants.M_SUN / constants.M_JUP
                )  # (Msun) -> (Mjup)
                param[5] = (
                    param[5] * constants.R_SUN / constants.R_JUP
                )  # (Rsun) -> (Rjup)

                iso_data.append(
                    [param[0], param[1], param[2], param[3], param[4], param[5]]
                )

            print(f"Adding isochrones: Sonora {labels[i]}...", end="", flush=True)

            iso_data = np.array(iso_data)

            metallicity = labels[i].split(" ")[2]

            dset = database.create_dataset(
                f"isochrones/sonora{metallicity}/age", data=iso_data[:, 0]
            )  # (Myr)
            database.create_dataset(
                f"isochrones/sonora{metallicity}/mass", data=iso_data[:, 1]
            )  # (Mjup)
            database.create_dataset(
                f"isochrones/sonora{metallicity}/log_lum", data=iso_data[:, 2]
            )  # log(L/Lsun)
            database.create_dataset(
                f"isochrones/sonora{metallicity}/teff", data=iso_data[:, 3]
            )  # (K)
            database.create_dataset(
                f"isochrones/sonora{metallicity}/log_g", data=iso_data[:, 4]
            )  # log(g)
            database.create_dataset(
                f"isochrones/sonora{metallicity}/radius", data=iso_data[:, 5]
            )  # (Rjup)

            dset.attrs["model"] = "sonora"

            print(" [DONE]")
            print(f"Database tag: sonora{metallicity}")



import os
import urllib.request

from species.data.isochrone_data.iso_manual import add_manual


def add_ames(database, input_path):
    """
    Function for adding the AMES-Cond and AMES-Dusty
    isochrone data to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    url_list = [
        "https://home.strw.leidenuniv.nl/~stolker/species/"
        "model.AMES-Cond-2000.M-0.0.MKO.Vega",
        "https://home.strw.leidenuniv.nl/~stolker/species/"
        "model.AMES-dusty.M-0.0.MKO.Vega",
    ]

    iso_tags = ["AMES-Cond", "AMES-Dusty"]
    iso_size = ["235 kB", "182 kB"]

    for i, url_item in enumerate(url_list):
        input_file = url_item.split("/")[-1]
        data_file = os.path.join(input_path, input_file)

        if not os.path.isfile(data_file):
            print(
                f"Downloading {iso_tags[i]} isochrones ({iso_size[i]})...",
                end="",
                flush=True,
            )
            urllib.request.urlretrieve(url_item, data_file)
            print(" [DONE]")

        add_manual(
            database=database,
            tag=iso_tags[i].lower(),
            file_name=data_file,
            model_name="ames",
        )



import numpy as np

from species.core import constants


def add_marleau(database, tag, file_name):
    """
    Function for adding the Marleau et al. isochrone data
    to the database. The isochrone data can be requested
    from Gabriel Marleau.

    https://ui.adsabs.harvard.edu/abs/2019A%26A...624A..20M/abstract

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    tag : str
        Tag name in the database.
    file_name : str
        Filename with the isochrones data.

    Returns
    -------
    NoneType
        None
    """

    # M      age     S_0             L          S(t)            R        Teff
    # (M_J)  (Gyr)   (k_B/baryon)    (L_sol)    (k_B/baryon)    (R_J)    (K)
    mass, age, _, luminosity, _, radius, teff = np.loadtxt(file_name, unpack=True)

    age *= 1e3  # (Myr)
    luminosity = np.log10(luminosity)

    mass_cgs = 1e3 * mass * constants.M_JUP  # (g)
    radius_cgs = 1e2 * radius * constants.R_JUP  # (cm)

    logg = np.log10(1e3 * constants.GRAVITY * mass_cgs / radius_cgs**2)

    print(f"Adding isochrones: {tag}...", end="", flush=True)

    isochrones = np.vstack((age, mass, teff, luminosity, logg))
    isochrones = np.transpose(isochrones)

    index_sort = np.argsort(isochrones[:, 0])
    isochrones = isochrones[index_sort, :]

    dset = database.create_dataset(f"isochrones/{tag}/evolution", data=isochrones)

    dset.attrs["model"] = "marleau"

    print(" [DONE]")



"""
Utility functions for ``petitRADTRANS`` spectra.
"""

from typing import Dict, List, Optional

import numpy as np

from typeguard import typechecked

from species.core.box import ModelBox
from species.read.read_radtrans import ReadRadtrans


@typechecked
def retrieval_spectrum(
    indices: Dict[str, np.int64],
    chemistry: str,
    pt_profile: str,
    line_species: List[str],
    cloud_species: List[str],
    quenching: Optional[str],
    spec_res: float,
    distance: Optional[float],
    pt_smooth: Optional[float],
    temp_nodes: Optional[np.integer],
    abund_nodes: Optional[np.integer],
    abund_smooth: Optional[float],
    read_rad: ReadRadtrans,
    sample: np.ndarray,
) -> ModelBox:
    """
    Function for calculating a petitRADTRANS spectrum
    from a posterior sample.

    Parameters
    ----------
    indices : dict
        Dictionary with the parameter indices for ``sample``.
    chemistry : str
        Chemistry type (``'equilibrium'`` or ``'free'``).
    pt_profile : str
        Pressure-temperature parametrization (``'molliere'``,
        ``'monotonic'``, or ``'free'``).
    line_species : list(str)
        List with the line species.
    cloud_species : list(str)
        List with the cloud species.
    quenching : str, None
        Quenching type for CO/CH4/H2O abundances. Either the quenching
        pressure (bar) is a free parameter (``quenching='pressure'``)
        or the quenching pressure is calculated from the mixing and
        chemical timescales (``quenching='diffusion'``). The quenching
        is not applied if the argument is set to ``None``.
    spec_res : float
        Spectral resolution.
    distance : float, None
        Distance (pc).
    pt_smooth : float, None
        Standard deviation of the Gaussian kernel that is used for
        smoothing the sampled temperature nodes of the P-T profile.
        Only required with `pt_profile='free'` or
        `pt_profile='monotonic'`. The argument should be given as
        log10(P/bar).
    temp_nodes : int, None
        Number of free temperature nodes that are used when
        ``pt_profile='monotonic'`` or ``pt_profile='free'``.
    abund_nodes : int, None
        Number of free abundance nodes that are used when
        ``chemistry='free'``.
    abund_smooth : float, None
        Standard deviation of the Gaussian kernel that is used for
        smoothing the abundance profiles, after the abundance nodes
        have been interpolated to a higher pressure resolution.
        Only required with ```chemistry='free'```. The argument
        should be given as :math:`\\log10{P/\\mathrm{bar}}`. No
        smoothing is applied if the argument if set to 0 or ``None``.
    read_rad : ReadRadtrans
        Instance of :class:`~species.read.read_radtrans.ReadRadtrans`.
    sample : np.ndarray
        Parameter values with their order given by the ``indices``.

    Returns
    -------
    species.core.box.ModelBox
        Box with the petitRADTRANS spectrum.
    """

    # Initiate parameter dictionary

    model_param = {}

    # Add log(g) and radius

    model_param["logg"] = sample[indices["logg"]]
    model_param["radius"] = sample[indices["radius"]]

    # Add distance

    if distance is not None:
        model_param["distance"] = distance

    # Add P-T profile parameters

    if pt_profile == "molliere":
        model_param["t1"] = sample[indices["t1"]]
        model_param["t2"] = sample[indices["t2"]]
        model_param["t3"] = sample[indices["t3"]]
        model_param["log_delta"] = sample[indices["log_delta"]]
        model_param["alpha"] = sample[indices["alpha"]]
        model_param["tint"] = sample[indices["tint"]]

    elif pt_profile == "gradient":
        num_layer = 6  # could make a variable in the future
        for index in range(num_layer):
            model_param[f"PTslope_{num_layer - index}"] = sample[
                indices[f"PTslope_{num_layer - index}"]
            ]
        model_param["T_bottom"] = sample[indices["T_bottom"]]

    elif pt_profile == "eddington":
        model_param["log_delta"] = sample[indices["log_delta"]]
        model_param["tint"] = sample[indices["tint"]]

    elif pt_profile in ["free", "monotonic"]:
        if temp_nodes is None:
            # For backward compatibility
            temp_nodes = 15

        for j in range(temp_nodes):
            model_param[f"t{j}"] = sample[indices[f"t{j}"]]

    if pt_smooth is not None:
        model_param["pt_smooth"] = pt_smooth

    # Add chemistry parameters

    if chemistry == "equilibrium":
        model_param["c_o_ratio"] = sample[indices["c_o_ratio"]]
        model_param["metallicity"] = sample[indices["metallicity"]]

    elif chemistry == "free":
        if abund_nodes is None:
            for line_item in line_species:
                model_param[line_item] = sample[indices[line_item]]

        else:
            for line_item in line_species:
                for node_idx in range(abund_nodes):
                    model_param[f"{line_item}_{node_idx}"] = sample[
                        indices[f"{line_item}_{node_idx}"]
                    ]

    if abund_smooth is not None:
        model_param["abund_smooth"] = abund_smooth

    if quenching == "pressure":
        model_param["log_p_quench"] = sample[indices["log_p_quench"]]

    # Add cloud parameters

    if "log_kappa_0" in indices:
        model_param["log_kappa_0"] = sample[indices["log_kappa_0"]]
        model_param["opa_index"] = sample[indices["opa_index"]]
        model_param["log_p_base"] = sample[indices["log_p_base"]]
        model_param["albedo"] = sample[indices["albedo"]]

        if "fsed" in indices:
            model_param["fsed"] = sample[indices["fsed"]]

        elif "fsed_1" in indices and "fsed_2" in indices:
            model_param["fsed_1"] = sample[indices["fsed_1"]]
            model_param["fsed_2"] = sample[indices["fsed_2"]]
            model_param["f_clouds"] = sample[indices["f_clouds"]]

        if "opa_knee" in indices:
            model_param["opa_knee"] = sample[indices["opa_knee"]]

    elif "log_kappa_abs" in indices:
        model_param["log_kappa_abs"] = sample[indices["log_kappa_abs"]]
        model_param["opa_abs_index"] = sample[indices["opa_abs_index"]]
        model_param["log_p_base"] = sample[indices["log_p_base"]]
        model_param["fsed"] = sample[indices["fsed"]]

        if "log_kappa_sca" in indices:
            model_param["log_kappa_sca"] = sample[indices["log_kappa_sca"]]
            model_param["opa_sca_index"] = sample[indices["opa_sca_index"]]
            model_param["lambda_ray"] = sample[indices["lambda_ray"]]

    elif "log_kappa_gray" in indices:
        model_param["log_kappa_gray"] = sample[indices["log_kappa_gray"]]
        model_param["log_cloud_top"] = sample[indices["log_cloud_top"]]

        if "albedo" in indices:
            model_param["albedo"] = sample[indices["albedo"]]

    elif len(cloud_species) > 0:
        model_param["fsed"] = sample[indices["fsed"]]
        model_param["sigma_lnorm"] = sample[indices["sigma_lnorm"]]

        if "kzz" in indices:
            # Backward compatibility
            model_param["kzz"] = sample[indices["kzz"]]

        elif "log_kzz" in indices:
            model_param["log_kzz"] = sample[indices["log_kzz"]]

        for cloud_item in cloud_species:
            cloud_param = f"{cloud_item[:-3].lower()}_fraction"

            if cloud_param in indices:
                model_param[cloud_param] = sample[indices[cloud_param]]

            cloud_param = f"{cloud_item[:-3].lower()}_tau"

            if cloud_param in indices:
                model_param[cloud_param] = sample[indices[cloud_param]]

            if cloud_item in indices:
                model_param[cloud_item] = sample[indices[cloud_item]]

    if "log_tau_cloud" in indices:
        model_param["tau_cloud"] = 10.0 ** sample[indices["log_tau_cloud"]]

        if len(cloud_species) > 1:
            for cloud_item in cloud_species[1:]:
                cloud_1 = cloud_item[:-3].lower()
                cloud_2 = cloud_species[0][:-3].lower()

                cloud_ratio = f"{cloud_1}_{cloud_2}_ratio"

                model_param[cloud_ratio] = sample[indices[cloud_ratio]]

    # Add extinction parameters

    if "ism_ext" in indices:
        model_param["ism_ext"] = sample[indices["ism_ext"]]

    if "ism_red" in indices:
        model_param["ism_red"] = sample[indices["ism_red"]]

    # Calculate spectrum

    model_box = read_rad.get_model(model_param, spec_res=spec_res)

    # Set content type of the ModelBox

    model_box.type = "mcmc"

    return model_box



"""
Module for adding the IRTF Spectral Library to the database.
"""

import os
import urllib.request

from typing import Optional, List

import h5py
import numpy as np
import pandas as pd

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import extract_tarfile, update_sptype
from species.util.query_util import get_parallax, get_simbad


@typechecked
def add_irtf(
    input_path: str, database: h5py._hl.files.File, sptypes: Optional[List[str]] = None
) -> None:
    """
    Function for adding the IRTF Spectral Library to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        Database.
    sptypes : list(str), None
        List with the spectral types ('F', 'G', 'K', 'M', 'L', 'T'). All spectral types are
        included if set to ``None``.

    Returns
    -------
    NoneType
        None
    """

    if sptypes is None:
        sptypes = ["F", "G", "K", "M", "L", "T"]

    parallax_url = "https://home.strw.leidenuniv.nl/~stolker/species/parallax.dat"
    parallax_file = os.path.join(input_path, "parallax.dat")

    if not os.path.isfile(parallax_file):
        urllib.request.urlretrieve(parallax_url, parallax_file)

    parallax_data = pd.pandas.read_csv(
        parallax_file,
        usecols=[0, 1, 2],
        names=["object", "parallax", "parallax_error"],
        delimiter=",",
        dtype={"object": str, "parallax": float, "parallax_error": float},
    )

    data_folder = os.path.join(input_path, "irtf")

    if not os.path.exists(data_folder):
        os.makedirs(data_folder)

    data_file = {
        "F": os.path.join(input_path, "irtf/F_fits_091201.tar"),
        "G": os.path.join(input_path, "irtf/G_fits_091201.tar"),
        "K": os.path.join(input_path, "irtf/K_fits_091201.tar"),
        "M": os.path.join(input_path, "irtf/M_fits_091201.tar"),
        "L": os.path.join(input_path, "irtf/L_fits_091201.tar"),
        "T": os.path.join(input_path, "irtf/T_fits_091201.tar"),
    }

    data_folder = {
        "F": os.path.join(input_path, "irtf/F_fits_091201"),
        "G": os.path.join(input_path, "irtf/G_fits_091201"),
        "K": os.path.join(input_path, "irtf/K_fits_091201"),
        "M": os.path.join(input_path, "irtf/M_fits_091201"),
        "L": os.path.join(input_path, "irtf/L_fits_091201"),
        "T": os.path.join(input_path, "irtf/T_fits_091201"),
    }

    main_folder = os.path.join(input_path, "irtf/")

    data_type = {
        "F": "F stars (4.4 MB)",
        "G": "G stars (5.6 MB)",
        "K": "K stars (5.5 MB)",
        "M": "M stars (7.5 MB)",
        "L": "L dwarfs (850 kB)",
        "T": "T dwarfs (100 kB)",
    }

    url_root = "http://irtfweb.ifa.hawaii.edu/~spex/IRTF_Spectral_Library/Data/"

    url = {
        "F": url_root + "F_fits_091201.tar",
        "G": url_root + "G_fits_091201.tar",
        "K": url_root + "K_fits_091201.tar",
        "M": url_root + "M_fits_091201.tar",
        "L": url_root + "L_fits_091201.tar",
        "T": url_root + "T_fits_091201.tar",
    }

    for item in sptypes:
        if not os.path.isfile(data_file[item]):
            print(
                f"Downloading IRTF Spectral Library - {data_type[item]}...",
                end="",
                flush=True,
            )
            urllib.request.urlretrieve(url[item], data_file[item])
            print(" [DONE]")

    print("Unpacking IRTF Spectral Library...", end="", flush=True)

    for item in sptypes:
        extract_tarfile(data_file[item], main_folder)

    print(" [DONE]")

    database.create_group("spectra/irtf")

    print_message = ""

    for item in sptypes:
        for root, _, files in os.walk(data_folder[item]):
            for _, filename in enumerate(files):
                if filename[-9:] != "_ext.fits":
                    fitsfile = os.path.join(root, filename)

                    spdata, header = fits.getdata(fitsfile, header=True)
                    spdata = np.transpose(spdata)

                    name = header["OBJECT"]
                    sptype = header["SPTYPE"]

                    if name[-2:] == "AB":
                        name = name[:-2]
                    elif name[-3:] == "ABC":
                        name = name[:-3]

                    spt_split = sptype.split()

                    if item in ["L", "T"] or spt_split[1][0] == "V":
                        empty_message = len(print_message) * " "
                        print(f"\r{empty_message}", end="")

                        print_message = f"Adding spectra... {name}"
                        print(f"\r{print_message}", end="")

                        simbad_id = get_simbad(name)

                        if simbad_id is not None:
                            # For backward compatibility
                            if not isinstance(simbad_id, str):
                                simbad_id = simbad_id.decode("utf-8")

                            par_select = parallax_data[
                                parallax_data["object"] == simbad_id
                            ]

                            if not par_select.empty:
                                parallax = (
                                    par_select["parallax"],
                                    par_select["parallax_error"],
                                )
                            else:
                                simbad_id, parallax = get_parallax(name)

                        else:
                            parallax = (np.nan, np.nan)

                        sptype = update_sptype(np.array([sptype]))[0]

                        dset = database.create_dataset(
                            f"spectra/irtf/{name}", data=spdata
                        )

                        dset.attrs["name"] = str(name).encode()
                        dset.attrs["sptype"] = str(sptype).encode()
                        dset.attrs["simbad"] = str(simbad_id).encode()
                        dset.attrs["parallax"] = parallax[0]
                        dset.attrs["parallax_error"] = parallax[1]

    empty_message = len(print_message) * " "
    print(f"\r{empty_message}", end="")

    print_message = "Adding spectra... [DONE]"
    print(f"\r{print_message}")



"""
Module for adding young, M- and L-type dwarf spectra from
`Bonnefoy et al. (2014) <https://ui.adsabs.harvard.edu/abs/
2014A%26A...562A.127B/abstract>`_ to the database.
"""

import gzip
import os
import shutil
import urllib.request

import h5py
import numpy as np

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import extract_tarfile


@typechecked
def add_bonnefoy2014(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the SINFONI spectra of young, M- and L-type
    dwarfs from `Bonnefoy et al. (2014) <https://ui.adsabs.harvard.
    edu/abs/2014A%26A...562A.127B/abstract>`_ to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        The HDF5 database.

    Returns
    -------
    NoneType
        None
    """

    print_text = "spectra of young M/L type objects from Bonnefoy et al. 2014"

    data_url = "http://cdsarc.u-strasbg.fr/viz-bin/nph-Cat/tar.gz?J/A+A/562/A127/"
    data_file = os.path.join(input_path, "J_A+A_562_A127.tar.gz")
    data_folder = os.path.join(input_path, "bonnefoy+2014/")

    if not os.path.isfile(data_file):
        print(f"Downloading {print_text} (2.3 MB)...", end="", flush=True)
        urllib.request.urlretrieve(data_url, data_file)
        print(" [DONE]")

    if os.path.exists(data_folder):
        shutil.rmtree(data_folder)

    print(f"Unpacking {print_text} (2.3 MB)...", end="", flush=True)
    extract_tarfile(data_file, data_folder)
    print(" [DONE]")

    spec_dict = {}

    with gzip.open(os.path.join(data_folder, "stars.dat.gz"), "r") as gzip_file:
        for line in gzip_file:
            name = line[:13].decode().strip()
            files = line[80:].decode().strip().split()
            sptype = line[49:56].decode().strip()

            if name == "NAME 2M1207A":
                name = "2M1207A"

            if len(sptype) == 0:
                sptype = None
            elif "." in sptype:
                sptype = sptype[:4]
            else:
                sptype = sptype[:2]

            if name == "Cha1109":
                sptype = "M9"
            elif name == "DH Tau B":
                sptype = "M9"
            elif name == "TWA 22A":
                sptype = "M6"
            elif name == "TWA 22B":
                sptype = "M6"
            elif name == "CT Cha b":
                sptype = "M9"

            spec_dict[name] = {"name": name, "sptype": sptype, "files": files}

    database.create_group("spectra/bonnefoy+2014")

    fits_folder = os.path.join(data_folder, "sp")

    print_message = ""

    for _, _, files in os.walk(fits_folder):
        for _, filename in enumerate(files):
            fname_split = filename.split("_")

            data = fits.getdata(os.path.join(fits_folder, filename))

            for name, value in spec_dict.items():
                if filename in value["files"]:
                    if name == "TWA 22AB":
                        # Binary spectrum
                        continue

                    if "JHK.fits" in fname_split:
                        value["JHK"] = data

                    elif "J" in fname_split:
                        value["J"] = data

                    elif "H+K" in fname_split or "HK" in fname_split:
                        value["HK"] = data

    for name, value in spec_dict.items():
        empty_message = len(print_message) * " "
        print(f"\r{empty_message}", end="")

        print_message = f"Adding spectra... {name}"
        print(f"\r{print_message}", end="")

        if "JHK" in value:
            sp_data = value["JHK"]

        elif "J" in value and "HK" in value:
            sp_data = np.vstack((value["J"], value["HK"]))

        else:
            continue

        dset = database.create_dataset(f"spectra/bonnefoy+2014/{name}", data=sp_data)

        dset.attrs["name"] = str(name).encode()
        dset.attrs["sptype"] = str(value["sptype"]).encode()

    empty_message = len(print_message) * " "
    print(f"\r{empty_message}", end="")

    print_message = "Adding spectra... [DONE]"
    print(f"\r{print_message}")



"""
Module with reading functionalities for filter profiles.
"""

import os
import warnings

from configparser import ConfigParser
from typing import Union, Tuple

import h5py
import numpy as np

from typeguard import typechecked
from scipy import interpolate

from species.data.filter_data.filter_data import add_filter_profile
from species.data.spec_data.spec_vega import add_vega


class ReadFilter:
    """
    Class for reading a filter profile from the database.
    """

    @typechecked
    def __init__(self, filter_name: str) -> None:
        """
        Parameters
        ----------
        filter_name : str
            Filter name as stored in the database. Filter names from
            the SVO Filter Profile Service will be automatically
            downloaded, stored in the database, and read from the
            database.

        Returns
        -------
        NoneType
            None
        """

        self.filter_name = filter_name

        config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.data_folder = config["species"]["data_folder"]

        with h5py.File(self.database, "r") as hdf5_file:
            # Check if the filter is found in 'r' mode
            # because the 'a' mode is not possible when
            # using multiprocessing
            if f"filters/{self.filter_name}" in hdf5_file:
                filter_found = True
            else:
                filter_found = False

        if not filter_found:
            with h5py.File(self.database, "a") as hdf5_file:
                add_filter_profile(self.data_folder, hdf5_file, self.filter_name)

    @typechecked
    def get_filter(self) -> np.ndarray:
        """
        Select a filter profile from the database.

        Returns
        -------
        np.ndarray
            Array with the wavelengths and filter transmission. The
            array has 2 dimensions with the shape (n_wavelengths, 2).
        """

        with h5py.File(self.database, "r") as hdf5_file:
            data = np.asarray(hdf5_file[f"filters/{self.filter_name}"])

            if data.shape[0] == 2 and data.shape[1] > data.shape[0]:
                # Required for backward compatibility
                data = np.transpose(data)

        return data

    @typechecked
    def interpolate_filter(self) -> interpolate.interp1d:
        """
        Interpolate a filter profile with the `interp1d <https://
        docs.scipy.org/doc/scipy/reference/generated/
        scipy.interpolate.interp1d.html>`_ function from
        ``scipy.interpolate`` and linear kind of interpolation.

        Returns
        -------
        scipy.interpolate.interp1d
            Linearly interpolated filter profile.
        """

        data = self.get_filter()

        return interpolate.interp1d(
            data[:, 0],
            data[:, 1],
            kind="linear",
            bounds_error=False,
            fill_value=float("nan"),
        )

    @typechecked
    def wavelength_range(
        self,
    ) -> Tuple[Union[np.float32, np.float64], Union[np.float32, np.float64]]:
        """
        Extract the wavelength range of the filter profile.

        Returns
        -------
        float
            Minimum wavelength (:math:`\\mu\\mathrm{m}`).
        float
            Maximum wavelength (:math:`\\mu\\mathrm{m}`).
        """

        data = self.get_filter()

        return data[0, 0], data[-1, 0]

    @typechecked
    def mean_wavelength(self) -> Union[np.float32, np.float64]:
        """
        Calculate the weighted mean wavelength of the filter profile.

        Returns
        -------
        float
            Mean wavelength (:math:`\\mu\\mathrm{m}`).
        """

        data = self.get_filter()

        return np.trapz(data[:, 0] * data[:, 1], x=data[:, 0]) / np.trapz(
            data[:, 1], x=data[:, 0]
        )

    @typechecked
    def effective_wavelength(self) -> Union[np.float32, np.float64]:
        """
        Calculate the effective wavelength of the filter profile.
        The effective wavelength is calculated as the weighted
        average based on the filter profile and the spectrum of Vega.

        Returns
        -------
        float
            Effective wavelength (:math:`\\mu\\mathrm{m}`).
        """

        filter_profile = self.get_filter()

        with h5py.File(self.database, "a") as hdf5_file:
            if "spectra/calibration/vega" not in hdf5_file:
                add_vega(self.data_folder, hdf5_file)

            vega_spec = np.array(hdf5_file["spectra/calibration/vega"])

        flux_interp = interpolate.interp1d(
            vega_spec[0,],
            vega_spec[1,],
            bounds_error=False,
            fill_value="extrapolate",
        )

        flux_filter = flux_interp(filter_profile[:, 0])

        return np.trapz(
            filter_profile[:, 0] * filter_profile[:, 1] * flux_filter,
            x=filter_profile[:, 0],
        ) / np.trapz(filter_profile[:, 1] * flux_filter, x=filter_profile[:, 0])

    @typechecked
    def filter_fwhm(self) -> float:
        """
        Calculate the full width at half maximum (FWHM) of the filter
        profile.

        Returns
        -------
        float
            Full width at half maximum (:math:`\\mu\\mathrm{m}`).
        """

        data = self.get_filter()

        spline = interpolate.InterpolatedUnivariateSpline(
            data[:, 0], data[:, 1] - np.max(data[:, 1]) / 2.0
        )
        root = spline.roots()

        diff = root - self.mean_wavelength()

        root1 = np.amax(diff[diff < 0.0])
        root2 = np.amin(diff[diff > 0.0])

        return root2 - root1

    @typechecked
    def effective_width(self) -> np.float32:
        """
        Calculate the effective width of the filter profile. The
        effective width is equivalent to the horizontal size of a
        rectangle with height equal to the maximum transmission and
        with the same area as the one covered by the filter profile.

        Returns
        -------
        float
            Effective width (:math:`\\mu\\mathrm{m}`).
        """

        data = self.get_filter()

        return np.trapz(data[:, 1], x=data[:, 0]) / np.amax(data[:, 1])

    @typechecked
    def detector_type(self) -> str:
        """
        Return the detector type.

        Returns
        -------
        str
            Detector type ('energy' or 'photon').
        """

        with h5py.File(self.database, "r") as hdf5_file:
            dset = hdf5_file[f"filters/{self.filter_name}"]

            if "det_type" in dset.attrs:
                det_type = dset.attrs["det_type"]

            else:
                warnings.warn(
                    f"Detector type not found for {self.filter_name}. "
                    "The database was probably created before the "
                    "detector type was introduced in species v0.3.1. "
                    "Assuming a photon-counting detector."
                )

                det_type = "photon"

        return det_type



"""
Module for the accretion luminosity relation.
"""

import os

import h5py
import numpy as np
import pooch

from typeguard import typechecked

from species.core import constants


@typechecked
def add_accretion_relation(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the accretion relation from `Aoyama et al.
    (2021) <https://ui. adsabs.harvard.edu/abs/
    2021ApJ...917L..30A/abstract>`_ and extrapolation from
    `Marleau & Aoyama (2022) <https://ui.adsabs.harvard.edu/abs/
    2022RNAAS...6..262M/abstract>`_ to the database. It provides
    coefficients to convert line to accretion luminosities.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    None
        NoneType
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    url = (
        "https://home.strw.leidenuniv.nl/~stolker/"
        "species/ab-Koeffienzenten_mehrStellen.dat"
    )

    data_file = os.path.join(input_path, "ab-Koeffienzenten_mehrStellen.dat")

    if not os.path.isfile(data_file):
        pooch.retrieve(
            url=url,
            known_hash="941b416e678128648c9ce485016af908f16bfe16ab72e0f8cb57a6bad963429a",
            fname="ab-Koeffienzenten_mehrStellen.dat",
            path=input_path,
            progressbar=True,
        )

    print("Adding coefficients for accretion relation (2.1 kB)...", end="", flush=True)

    data = np.genfromtxt(
        data_file,
        dtype=None,
        skip_header=1,
        encoding=None,
        names=True,
        usecols=[0, 1, 2, 3, 4],
    )

    line_names = data["name"]
    coefficients = np.column_stack([data["a"], data["b"]])

    n_init = data["ni"]
    n_final = data["nf"]

    delta_n_min = {"H": 9, "Pa": 8, "Br": 6}
    delta_n_max = {"H": 14, "Pa": 13, "Br": 12}

    for i, item in enumerate(["H", "Pa", "Br"]):
        for j in range(delta_n_min[item], delta_n_max[item] + 1):
            # n_f: Ly=1, H=Ba=2, Pa=3, Br=4
            n_f_tmp = i + 2

            idx_insert = np.argwhere(line_names == f"{item}{j+n_f_tmp-1}")[0][0] + 1
            line_names = np.insert(line_names, idx_insert, f"{item}{j+n_f_tmp}")
            n_final = np.insert(n_final, idx_insert, n_f_tmp)

            # delta_n = n_i - n_f
            n_i_tmp = j + n_f_tmp
            n_init = np.insert(n_init, idx_insert, n_i_tmp)

            # Relation for extrapolation of coefficients
            # See Marleau & Aoyama (2022)

            a_coeff = 0.811 - (1.0 / (9.90 * n_f_tmp - 9.5 * n_i_tmp))

            b_coeff = (
                1.0
                + 1.05 * np.log(n_i_tmp)
                + (1.0 / (n_i_tmp - n_f_tmp))
                - (1.0 / n_f_tmp)
            ) * (1.07 + 0.0694 * n_f_tmp) - 1.41

            coefficients = np.insert(
                coefficients, idx_insert, [[a_coeff, b_coeff]], axis=0
            )

    # Rest vacuum wavelength (um) from Rydberg formula,
    # which is valid usually to three or four decimal places.
    # Exact values, if needed, can be obtained from e.g.
    # Wiese & Fuhr (2009):
    # http://adsabs.harvard.edu/abs/2009JPCRD..38..565W
    # https://www.nist.gov/system/files/documents/srd/jpcrd382009565p.pdf
    wavelengths = 1e6 / (constants.RYDBERG * (1.0 / n_final**2 - 1.0 / n_init**2))

    # data = np.column_stack(
    #     [
    #         line_names,
    #         wavelengths,
    #         n_init,
    #         n_final,
    #         coefficients[:, 0],
    #         coefficients[:, 1],
    #     ]
    # )
    # np.savetxt("acc_lines.dat", data, delimiter=" ", fmt="%s")

    database.create_dataset("accretion/wavelengths", data=wavelengths)
    database.create_dataset("accretion/coefficients", data=coefficients)

    dtype = h5py.special_dtype(vlen=str)
    dset = database.create_dataset(
        "accretion/hydrogen_lines", (np.size(line_names),), dtype=dtype
    )
    dset[...] = line_names

    print(" [DONE]")

    print(
        "Please cite Aoyama et al. (2021) and Marleau & Aoyama "
        "(2022) when using the accretion relation in a publication"
    )



"""
Module with functionalities for comparing a spectrum with a library of
empirical or model spectra. Empirical libraries of field or low-gravity
objects can be used to determine the spectral type. A comparison with
model spectra are useful for exploration of the atmospheric parameter.
"""

import configparser
import os
import warnings

from typing import List, Optional, Tuple, Union

import h5py
import numpy as np

from scipy.interpolate import interp1d
from typeguard import typechecked

from species.core import constants
from species.data.spec_data.add_spec_data import add_spec_library
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.read.read_model import ReadModel
from species.read.read_object import ReadObject
from species.util.dust_util import ism_extinction
from species.util.spec_util import smooth_spectrum


class CompareSpectra:
    """
    Class for comparing the spectrum of an object with
    a library of empirical or model spectra.
    """

    @typechecked
    def __init__(
        self,
        object_name: str,
        spec_name: Union[str, List[str]],
    ) -> None:
        """
        Parameters
        ----------
        object_name : str
            Object name as stored in the database with
            :func:`~species.data.database.Database.add_object` or
            :func:`~species.data.database.Database.add_companion`.
        spec_name : str, list(str)
            Name of the spectrum or a list with the names of the
            spectra that will be used for the comparison. The
            spectrum names should have been stored at the object
            data of ``object_name``.

        Returns
        -------
        NoneType
            None
        """

        self.object_name = object_name
        self.spec_name = spec_name

        if isinstance(self.spec_name, str):
            self.spec_name = [self.spec_name]

        self.object = ReadObject(object_name)

        config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = configparser.ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.data_folder = config["species"]["data_folder"]

    @typechecked
    def spectral_type(
        self,
        tag: str,
        spec_library: str,
        wavel_range: Optional[Tuple[Optional[float], Optional[float]]] = None,
        sptypes: Optional[List[str]] = None,
        av_ext: Optional[Union[List[float], np.ndarray]] = None,
        rad_vel: Optional[Union[List[float], np.ndarray]] = None,
    ) -> None:
        """
        Method for finding the best fitting empirical spectra
        from a selected library by evaluating the goodness-of-fit
        statistic from `Cushing et al. (2008) <https://ui.adsabs.
        harvard.edu/abs/2008ApJ...678.1372C/abstract>`_.

        Parameters
        ----------
        tag : str
            Database tag where for each spectrum from the spectral
            library the best-fit parameters will be stored. So when
            testing a range of values for ``av_ext`` and ``rad_vel``,
            only the parameters that minimize the goodness-of-fit
            statistic will be stored.
        spec_library : str
            Name of the spectral library (e.g. 'irtf', 'spex',
            'kesseli+2017', 'bonnefoy+2014').
        wavel_range : tuple(float, float), None
            Wavelength range (:math:`\\mu\\mathrm{m}`) that
            is evaluated.
        sptypes : list(str), None
            List with spectral types to compare with, for example
            ``sptypes=['M', 'L']``. All available spectral types
            in the ``spec_library`` are compared with if the
            argument is set to ``None``.
        av_ext : list(float), np.ndarray, None
            List of :math:`A_V` for which the goodness-of-fit
            statistic is tested. The extinction is calculated with
            the empirical relation from `Cardelli et al. (1989)
            <https://ui.adsabs.harvard.edu/abs/1989ApJ...345..245C/
            abstract>`_. The extinction is not varied if the argument
            is set to ``None``.
        rad_vel : list(float), np.ndarray, None
            List of radial velocities (km s-1) for which the
            goodness-of-fit statistic is tested. The radial
            velocity is not varied if the argument is set
            to ``None``.

        Returns
        -------
        NoneType
            None
        """

        w_i = 1.0

        if av_ext is None:
            av_ext = [0.0]

        if rad_vel is None:
            rad_vel = [0.0]

        with h5py.File(self.database, "a") as hdf5_file:
            if f"spectra/{spec_library}" not in hdf5_file:
                add_spec_library(self.data_folder, hdf5_file, spec_library)

        # Read object spectra and resolution

        obj_spec = []
        obj_res = []

        for item in self.spec_name:
            obj_spec.append(self.object.get_spectrum()[item][0])
            obj_res.append(self.object.get_spectrum()[item][3])

        # Read inverted covariance matrix
        # obj_inv_cov = self.object.get_spectrum()[self.spec_name][2]

        # Create empty lists for results

        name_list = []
        spt_list = []
        gk_list = []
        ck_list = []
        av_list = []
        rv_list = []

        print_message = ""

        # Start looping over library spectra

        for i, item in enumerate(hdf5_file[f"spectra/{spec_library}"]):
            # Read spectrum spectral type from library
            dset = hdf5_file[f"spectra/{spec_library}/{item}"]

            if isinstance(dset.attrs["sptype"], str):
                item_sptype = dset.attrs["sptype"]
            else:
                # Use decode for backward compatibility
                item_sptype = dset.attrs["sptype"].decode("utf-8")

            if item_sptype == "None":
                continue

            if sptypes is None or item_sptype[0] in sptypes:
                # Convert HDF5 dataset into numpy array
                spectrum = np.asarray(dset)

                if wavel_range is not None:
                    # Select subset of the spectrum

                    if wavel_range[0] is None:
                        indices = np.where((spectrum[:, 0] < wavel_range[1]))[0]

                    elif wavel_range[1] is None:
                        indices = np.where((spectrum[:, 0] > wavel_range[0]))[0]

                    else:
                        indices = np.where(
                            (spectrum[:, 0] > wavel_range[0])
                            & (spectrum[:, 0] < wavel_range[1])
                        )[0]

                    if len(indices) == 0:
                        raise ValueError(
                            "The selected wavelength range does not "
                            "cover any wavelength points of the input "
                            "spectrum. Please use a broader range as "
                            "argument of 'wavel_range'."
                        )

                    spectrum = spectrum[indices,]

                empty_message = len(print_message) * " "
                print(f"\r{empty_message}", end="")

                print_message = f"Processing spectra... {item}"
                print(f"\r{print_message}", end="")

                # Loop over all values of A_V and RV that will be tested

                for av_item in av_ext:
                    for rv_item in rad_vel:
                        for j, spec_item in enumerate(obj_spec):
                            # Dust extinction
                            ism_ext = ism_extinction(av_item, 3.1, spectrum[:, 0])
                            flux_scaling = 10.0 ** (-0.4 * ism_ext)

                            # Shift wavelengths by RV
                            wavel_shifted = (
                                spectrum[:, 0]
                                + spectrum[:, 0] * 1e3 * rv_item / constants.LIGHT
                            )

                            # Smooth spectrum
                            flux_smooth = smooth_spectrum(
                                wavel_shifted,
                                spectrum[:, 1] * flux_scaling,
                                spec_res=obj_res[j],
                                force_smooth=True,
                            )

                            # Interpolate library spectrum to object wavelengths
                            interp_spec = interp1d(
                                spectrum[:, 0],
                                flux_smooth,
                                kind="linear",
                                fill_value="extrapolate",
                            )

                            indices = np.where(
                                (spec_item[:, 0] > np.amin(spectrum[:, 0]))
                                & (spec_item[:, 0] < np.amax(spectrum[:, 0]))
                            )[0]

                            flux_resample = interp_spec(spec_item[indices, 0])

                            c_numer = (
                                w_i
                                * spec_item[indices, 1]
                                * flux_resample
                                / spec_item[indices, 2] ** 2
                            )

                            c_denom = (
                                w_i * flux_resample**2 / spec_item[indices, 2] ** 2
                            )

                            if j == 0:
                                g_k = 0.0
                                c_k_spec = []

                            c_k = np.sum(c_numer) / np.sum(c_denom)
                            c_k_spec.append(c_k)

                            chi_sq = (
                                spec_item[indices, 1] - c_k * flux_resample
                            ) / spec_item[indices, 2]

                            g_k += np.sum(w_i * chi_sq**2)

                            # obj_inv_cov_crop = obj_inv_cov[indices, :]
                            # obj_inv_cov_crop = obj_inv_cov_crop[:, indices]

                            # g_k = np.dot(spec_item[indices, 1]-c_k*flux_resample,
                            #     np.dot(obj_inv_cov_crop,
                            #            spec_item[indices, 1]-c_k*flux_resample))

                        # Append to the lists of results

                        name_list.append(item)
                        spt_list.append(item_sptype)
                        gk_list.append(g_k)
                        ck_list.append(c_k_spec)
                        av_list.append(av_item)
                        rv_list.append(rv_item)

        empty_message = len(print_message) * " "
        print(f"\r{empty_message}", end="")

        print("\rProcessing spectra... [DONE]")

        hdf5_file.close()

        name_list = np.asarray(name_list)
        spt_list = np.asarray(spt_list)
        gk_list = np.asarray(gk_list)
        ck_list = np.asarray(ck_list)
        av_list = np.asarray(av_list)
        rv_list = np.asarray(rv_list)

        sort_index = np.argsort(gk_list)

        name_list = name_list[sort_index]
        spt_list = spt_list[sort_index]
        gk_list = gk_list[sort_index]
        ck_list = ck_list[sort_index]
        av_list = av_list[sort_index]
        rv_list = rv_list[sort_index]

        name_select = []
        spt_select = []
        gk_select = []
        ck_select = []
        av_select = []
        rv_select = []

        for i, item in enumerate(name_list):
            if item not in name_select:
                name_select.append(item)
                spt_select.append(spt_list[i])
                gk_select.append(gk_list[i])
                ck_select.append(ck_list[i])
                av_select.append(av_list[i])
                rv_select.append(rv_list[i])

        print("Best-fitting spectra:")

        if len(gk_select) < 10:
            for i, gk_item in enumerate(gk_select):
                print(
                    f"   {i+1:2d}. G = {gk_item:.2e} -> {name_select[i]}, {spt_select[i]}, "
                    f"A_V = {av_select[i]:.2f}, RV = {rv_select[i]:.0f} km/s,\n"
                    f"                      scalings = {ck_select[i]}"
                )

        else:
            for i in range(10):
                print(
                    f"   {i+1:2d}. G = {gk_select[i]:.2e} -> {name_select[i]}, {spt_select[i]}, "
                    f"A_V = {av_select[i]:.2f}, RV = {rv_select[i]:.0f} km/s,\n"
                    f"                      scalings = {ck_select[i]}"
                )

        from species.data.database import Database

        species_db = Database()

        species_db.add_empirical(
            tag=tag,
            names=name_select,
            sptypes=spt_select,
            goodness_of_fit=gk_select,
            flux_scaling=ck_select,
            av_ext=av_select,
            rad_vel=rv_select,
            object_name=self.object_name,
            spec_name=self.spec_name,
            spec_library=spec_library,
        )

    @typechecked
    def compare_model(
        self,
        tag: str,
        model: str,
        av_points: Optional[Union[List[float], np.ndarray]] = None,
        fix_logg: Optional[float] = None,
        scale_spec: Optional[List[str]] = None,
        weights: bool = True,
        inc_phot: Union[List[str], bool] = False,
    ) -> None:
        """
        Method for finding the best-fit spectrum from a grid of
        atmospheric model spectra by evaluating the goodness-of-fit
        statistic from `Cushing et al. (2008) <https://ui.adsabs.
        harvard.edu/abs/2008ApJ...678.1372C/abstract>`_.

        Parameters
        ----------
        tag : str
            Database tag where for each spectrum from the spectral
            library the best-fit parameters will be stored.
        model : str
            Name of the model grid with synthetic spectra.
        av_points : list(float), np.ndarray, None
            List of :math:`A_V` extinction values for which the
            goodness-of-fit statistic will be tested. The extinction is
            calculated with the relation from `Cardelli et al. (1989)
            <https://ui.adsabs.harvard.edu/abs/1989ApJ...345..245C/
            abstract>`_.
        fix_logg : float, None
            Fix the value of :math:`\\log(g)`, for example if estimated
            from gravity-sensitive spectral features. Typically,
            :math:`\\log(g)` can not be accurately determined when
            comparing the spectra over a broad wavelength range.
        scale_spec : list(str), None
            List with names of observed spectra to which an additional
            flux scaling is applied to best match the spectral
            templates. This can be used to account for a difference in
            absolute calibration between spectra.
        weights : bool
            Apply a weighting of the spectra and photometry based on
            the widths of the wavelengths bins and the FWHM of the
            filter profiles, respectively. No weighting is applied
            if the argument is set to ``False``.
        inc_phot : list(str), bool
            Filter names for which photometric fluxes of the
            selected ``object_name`` will be included in the
            comparison. The argument can be a list with filter names
            or a boolean to select all or none of the photometry.
            By default the argument is set to ``False`` so
            photometric fluxes are not included.

        Returns
        -------
        NoneType
            None
        """

        w_i = {}

        for spec_item in self.spec_name:
            # Determine width of wavelength bins for the spectra
            # Set the optional weights for the statistic to the
            # width of the bins or otherwise to one
            obj_wavel = self.object.get_spectrum()[spec_item][0][:, 0]

            diff = (np.diff(obj_wavel)[1:] + np.diff(obj_wavel)[:-1]) / 2.0
            diff = np.insert(diff, 0, diff[0])
            diff = np.append(diff, diff[-1])

            if weights:
                w_i[spec_item] = diff
            else:
                w_i[spec_item] = np.ones(obj_wavel.shape[0])

        read_object = ReadObject(self.object_name)

        if isinstance(inc_phot, bool):
            # The argument of inc_phot is a boolean
            if inc_phot:
                # Select all filters if inc_phot=True
                inc_phot = read_object.list_filters()

            else:
                inc_phot = []

        object_flux = {}
        for filter_item in inc_phot:
            object_flux[filter_item] = read_object.get_photometry(filter_item)[2:]

        if scale_spec is None:
            scale_spec = []

        phot_wavel = {}
        for phot_item in inc_phot:
            read_filt = ReadFilter(phot_item)
            phot_wavel[phot_item] = read_filt.mean_wavelength()

            if weights:
                # Set the weight for the photometric fluxes
                # to the FWHM of the filter profile
                w_i[phot_item] = read_filt.filter_fwhm()
            else:
                w_i[phot_item] = 1.0

        model_reader = ReadModel(model)
        model_param = model_reader.get_parameters()
        grid_points = model_reader.get_points()

        coord_points = []
        for key, value in grid_points.items():
            if key == "logg" and fix_logg is not None:
                if fix_logg in value:
                    coord_points.append(np.array([fix_logg]))

                else:
                    raise ValueError(
                        f"The argument of 'fix_logg' ({fix_logg}) is "
                        f"not found in the parameter grid of the "
                        f"model spectra. The following values of "
                        f"log(g) are available: {value}"
                    )

            else:
                coord_points.append(value)

        if av_points is not None:
            model_param.append("ism_ext")
            coord_points.append(av_points)

        grid_shape = []
        for item in coord_points:
            grid_shape.append(len(item))

        for _ in range(len(coord_points), 6):
            model_param.append(None)
            coord_points.append([None])
            grid_shape.append(1)

        spec_data = self.object.get_spectrum()

        fit_stat = np.zeros(grid_shape)
        flux_scaling = np.zeros(grid_shape)

        if len(scale_spec) > 0:
            grid_shape.append(len(scale_spec))
            extra_scaling = np.zeros(grid_shape)

        else:
            extra_scaling = None

        n_iter = 1
        for item in coord_points:
            if len(item) > 0:
                n_iter *= len(item)

        count = 1

        for coord_0_idx, coord_0_item in enumerate(coord_points[0]):
            for coord_1_idx, coord_1_item in enumerate(coord_points[1]):
                for coord_2_idx, coord_2_item in enumerate(coord_points[2]):
                    for coord_3_idx, coord_3_item in enumerate(coord_points[3]):
                        for coord_4_idx, coord_4_item in enumerate(coord_points[4]):
                            for coord_5_idx, coord_5_item in enumerate(coord_points[5]):
                                print(
                                    f"\rProcessing model spectrum {count}/{n_iter}...",
                                    end="",
                                )

                                param_dict = {}
                                model_spec = {}

                                if model_param[0] is not None:
                                    param_dict[model_param[0]] = coord_0_item

                                if model_param[1] is not None:
                                    param_dict[model_param[1]] = coord_1_item

                                if model_param[2] is not None:
                                    param_dict[model_param[2]] = coord_2_item

                                if model_param[3] is not None:
                                    param_dict[model_param[3]] = coord_3_item

                                if model_param[4] is not None:
                                    param_dict[model_param[4]] = coord_4_item

                                if model_param[5] is not None:
                                    param_dict[model_param[5]] = coord_5_item

                                model_box_full = model_reader.get_data(param_dict)

                                for spec_item in self.spec_name:
                                    obj_spec = self.object.get_spectrum()[spec_item][0]
                                    obj_res = self.object.get_spectrum()[spec_item][3]

                                    # Smooth model spectrum

                                    model_flux = smooth_spectrum(
                                        model_box_full.wavelength,
                                        model_box_full.flux,
                                        obj_res,
                                    )

                                    # Resample model spectrum

                                    flux_intep = interp1d(
                                        model_box_full.wavelength,
                                        model_flux,
                                        bounds_error=False,
                                    )
                                    model_flux = flux_intep(obj_spec[:, 0])

                                    nan_wavel = np.sum(np.isnan(model_flux))

                                    if nan_wavel > 0:
                                        warnings.warn(
                                            "The extracted model spectrum contains "
                                            f"{nan_wavel} fluxes with NaN. Probably "
                                            "because some of the wavelengths of the "
                                            "data are outside the available "
                                            "wavelength range of the model grid. "
                                            "These wavelengths are ignored when "
                                            "calculating the goodness-of-fit statistic."
                                        )

                                    model_spec[spec_item] = model_flux

                                g_fit = 0.0

                                model_list = []
                                data_list = []
                                weights_list = []

                                for spec_item in self.spec_name:
                                    if spec_item not in scale_spec:
                                        model_list.append(model_spec[spec_item])
                                        data_list.append(spec_data[spec_item][0])
                                        weights_list.append(w_i[spec_item])

                                model_phot = {}

                                for phot_item in inc_phot:
                                    syn_phot = SyntheticPhotometry(phot_item)
                                    model_phot[phot_item] = syn_phot.spectrum_to_flux(
                                        model_box_full.wavelength, model_box_full.flux
                                    )[0]
                                    model_list.append(np.array([model_phot[phot_item]]))

                                    phot_flux = object_flux[phot_item]
                                    phot_data = np.array(
                                        [
                                            [
                                                phot_wavel[phot_item],
                                                phot_flux[0],
                                                phot_flux[1],
                                            ]
                                        ]
                                    )
                                    data_list.append(phot_data)

                                    weights_list.append(np.array([w_i[phot_item]]))

                                data_list = np.concatenate(data_list)
                                model_list = np.concatenate(model_list)
                                weights_list = np.concatenate(weights_list)

                                c_numer = (
                                    weights_list
                                    * data_list[:, 1]
                                    * model_list
                                    / data_list[:, 2] ** 2
                                )

                                c_denom = (
                                    weights_list
                                    * model_list**2
                                    / data_list[:, 2] ** 2
                                )

                                if np.nansum(model_list) == 0.0:
                                    # This happens if model spectra contain
                                    # only zeros because the grid point was
                                    # missing and could not be fixed with
                                    # an interpolation when the grid was
                                    # added to the database
                                    scaling = np.nan

                                else:
                                    scaling = np.nansum(c_numer) / np.nansum(c_denom)

                                flux_scaling[
                                    coord_0_idx,
                                    coord_1_idx,
                                    coord_2_idx,
                                    coord_3_idx,
                                    coord_4_idx,
                                    coord_5_idx,
                                ] = scaling

                                for spec_item in scale_spec:
                                    spec_idx = scale_spec.index(spec_item)

                                    c_numer = (
                                        w_i[spec_item]
                                        * scaling
                                        * model_spec[spec_item]
                                        * spec_data[spec_item][0][:, 1]
                                        / spec_data[spec_item][0][:, 2] ** 2
                                    )

                                    c_denom = (
                                        w_i[spec_item]
                                        * spec_data[spec_item][0][:, 1] ** 2
                                        / spec_data[spec_item][0][:, 2] ** 2
                                    )

                                    spec_scaling = np.nansum(c_numer) / np.nansum(
                                        c_denom
                                    )

                                    extra_scaling[
                                        coord_0_idx,
                                        coord_1_idx,
                                        coord_2_idx,
                                        coord_3_idx,
                                        coord_4_idx,
                                        coord_5_idx,
                                        spec_idx,
                                    ] = spec_scaling

                                for phot_item in inc_phot:
                                    phot_flux = object_flux[phot_item]

                                    g_fit += np.nansum(
                                        w_i[phot_item]
                                        * (
                                            phot_flux[0]
                                            - scaling * model_phot[phot_item]
                                        )
                                        ** 2
                                        / phot_flux[1] ** 2
                                    )

                                for spec_item in self.spec_name:
                                    if spec_item in scale_spec:
                                        spec_idx = scale_spec.index(spec_item)

                                        data_scaling = extra_scaling[
                                            coord_0_idx,
                                            coord_1_idx,
                                            coord_2_idx,
                                            coord_3_idx,
                                            coord_4_idx,
                                            coord_5_idx,
                                            spec_idx,
                                        ]

                                    else:
                                        data_scaling = 1.0

                                    g_fit += np.nansum(
                                        w_i[spec_item]
                                        * (
                                            data_scaling * spec_data[spec_item][0][:, 1]
                                            - scaling * model_spec[spec_item]
                                        )
                                        ** 2
                                        / spec_data[spec_item][0][:, 2] ** 2
                                    )

                                if np.nansum(model_list) == 0.0:
                                    # This happens if model spectra contain
                                    # only zeros because the grid point was
                                    # missing and could not be fixed with
                                    # an interpolation when the grid was
                                    # added to the database
                                    g_fit = np.nan

                                fit_stat[
                                    coord_0_idx,
                                    coord_1_idx,
                                    coord_2_idx,
                                    coord_3_idx,
                                    coord_4_idx,
                                    coord_5_idx,
                                ] = g_fit

                                count += 1

        print(" [DONE]")

        for param_idx, param_item in enumerate(model_param):
            if param_item is None:
                model_param = model_param[:param_idx]
                coord_points = coord_points[:param_idx]
                break

        for dim_idx in range(fit_stat.ndim, 0, -1):
            if dim_idx > len(model_param):
                fit_stat = fit_stat[..., 0]
                flux_scaling = flux_scaling[..., 0]

                if extra_scaling is not None:
                    extra_scaling = extra_scaling[..., 0, :]

        from species.data.database import Database

        species_db = Database()

        species_db.add_comparison(
            tag=tag,
            goodness_of_fit=fit_stat,
            flux_scaling=flux_scaling,
            model_param=model_param,
            coord_points=coord_points,
            object_name=self.object_name,
            spec_name=self.spec_name,
            model=model,
            scale_spec=scale_spec,
            extra_scaling=extra_scaling,
            inc_phot=inc_phot,
        )



"""
Module for adding a grid of model spectra to the database.
"""

import json
import os
import pathlib
import tarfile
import warnings

from typing import Optional, Tuple

import h5py
import numpy as np
import pooch
import spectres

from typeguard import typechecked

from species.util.core_util import print_section
from species.util.data_util import add_missing, extract_tarfile, sort_data, write_data
from species.util.model_util import convert_model_name
from species.util.spec_util import create_wavelengths


@typechecked
def add_model_grid(
    model_tag: str,
    input_path: str,
    database: h5py._hl.files.File,
    wavel_range: Optional[Tuple[float, float]] = None,
    teff_range: Optional[Tuple[float, float]] = None,
    wavel_sampling: Optional[float] = None,
    unpack_tar: bool = True,
) -> None:
    """
    Function for adding a grid of model spectra to the database.
    The original spectra had been resampled to logarithmically-
    spaced wavelengths, so with at a constant
    :math:`\\lambda/\\Delta\\lambda`. This function downloads
    the model grid, unpacks the tar file, and adds the spectra
    and parameters to the database.

    Parameters
    ----------
    model_tag : str
        Tag of the grid of model spectra.
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        HDF5 database.
    wavel_range : tuple(float, float), None
        Wavelength range (:math:`\\mu\\text{m}`). The original
        wavelength points are used if the argument is set to ``None``.
    teff_range : tuple(float, float), None
        Range of effective temperatures (K) for which the spectra will
        be extracted from the TAR file and added to the database. All
        spectra are selected if the argument is set to ``None``.
    wavel_sampling : float, None
        Wavelength spacing :math:`\\lambda/\\Delta\\lambda` to which
        the spectra will be resampled. Typically this parameter is
        not needed so the argument can be set to ``None``. The only
        benefit of using this parameter is limiting the storage
        in the HDF5 database. The parameter should be used in
        combination with setting the ``wavel_range``.
    unpack_tar : bool
        Unpack the TAR file with the model spectra in the
        ``data_folder``. The argument can be set to ``False`` if
        the TAR file had already been unpacked previously.

    Returns
    -------
    NoneType
        None
    """

    print_section("Add grid of model spectra")

    model_name = convert_model_name(model_tag)
    print(f"Database tag: {model_tag}")
    print(f"Model name: {model_name}")

    data_file = pathlib.Path(__file__).parent.resolve() / "model_data.json"

    with open(data_file, "r", encoding="utf-8") as json_file:
        model_data = json.load(json_file)

    if model_tag in model_data.keys():
        model_info = model_data[model_tag]

    else:
        raise ValueError(
            f"The '{model_tag}' atmospheric model is not available. "
            "Please choose one of the following models: "
            "'ames-cond', 'ames-dusty', 'atmo', 'bt-settl', "
            "'bt-nextgen', 'drift-phoexnix', 'petitcode-cool-clear', "
            "'petitcode-cool-cloudy', 'petitcode-hot-clear', "
            "'petitcode-hot-cloudy', 'exo-rem', 'bt-settl-cifist', "
            "'bt-cond', 'bt-cond-feh', 'blackbody', 'sonora-cholla', "
            "'sonora-bobcat', 'sonora-bobcat-co', 'koester-wd', "
            "'saumon2008-clear', 'saumon2008-cloudy', 'petrus2023', "
            "'sphinx'"
        )

    if model_tag == "bt-settl":
        warnings.warn(
            "It is recommended to use the CIFIST "
            "grid of the BT-Settl, because it is "
            "a newer version. In that case, set "
            "model='bt-settl-cifist' when using "
            "add_model of Database."
        )

    elif model_tag == "exo-rem":
        warnings.warn(
            "The Exo-Rem grid has been updated to the latest version "
            "from https://lesia.obspm.fr/exorem/YGP_grids/. Please "
            "consider removing the grid from the 'data_folder' if "
            "needed such that the latest version of the grid will "
            "be downloaded and added to the HDF5 database."
        )

    elif model_tag == "exo-rem-highres":
        if teff_range is None:
            warnings.warn(
                "Adding the full high-resolution grid of Exo-Rem to the "
                "HDF5 database may not be feasible since it requires "
                "a large amount of memory. Please consider using the "
                "'teff_range' parameter to only add a subset of the "
                "model spectra to the database."
            )

        if wavel_range is None:
            warnings.warn(
                "Adding the full high-resolution grid of Exo-Rem to the "
                "HDF5 database may not be feasible since it requires "
                "a large amount of memory. Please consider using the "
                "'wavel_range' parameter to reduce the data size."
            )

    if wavel_sampling is not None and wavel_range is None:
        warnings.warn(
            "The 'wavel_sampling' parameter can only be "
            "used in combination with the 'wavel_range' "
            "parameter. The model spectra are therefore "
            "not resampled."
        )

        wavel_sampling = None

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    input_file = f"{model_tag}.tgz"

    data_folder = os.path.join(input_path, model_tag)
    data_file = os.path.join(input_path, input_file)

    if not os.path.exists(data_folder):
        os.makedirs(data_folder)

    url = f"https://home.strw.leidenuniv.nl/~stolker/species/{model_tag}.tgz"

    if not os.path.isfile(data_file):
        pooch.retrieve(
            url=url,
            known_hash=None,
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    if unpack_tar:
        with tarfile.open(data_file) as tar_open:
            # Get a list of all TAR members
            tar_members = tar_open.getmembers()

        if teff_range is None:
            member_list = None
            n_members = len(tar_members)

        else:
            # Only include and extract TAR members
            # within the specified Teff range
            member_list = []

            for tar_item in tar_members:
                file_split = tar_item.name.split("_")
                param_index = file_split.index("teff") + 1
                teff_val = float(file_split[param_index])

                if teff_range[0] <= teff_val <= teff_range[1]:
                    member_list.append(tar_item)

            n_members = len(member_list)

        print(
            f"\nUnpacking {n_members}/{len(tar_members)} model spectra "
            f"from {model_info['name']} ({model_info['file size']})...",
            end="",
            flush=True,
        )

        extract_tarfile(data_file, data_folder, member_list=member_list)

        print(" [DONE]")

    print_newline = False

    if "information" in model_info:
        if not print_newline:
            print()
            print_newline = True

        print(f"Model information: {model_info['information']}")

    if "reference" in model_info:
        if not print_newline:
            print()
            print_newline = True

        print(
            f"Please cite {model_info['reference']} when "
            f"using {model_info['name']} in a publication"
        )

    if "url" in model_info:
        if not print_newline:
            print()
            print_newline = True

        print(f"Reference URL: {model_info['url']}")

    teff = []

    if "logg" in model_info["parameters"]:
        logg = []
    else:
        logg = None

    if "feh" in model_info["parameters"]:
        feh = []
    else:
        feh = None

    if "c_o_ratio" in model_info["parameters"]:
        c_o_ratio = []
    else:
        c_o_ratio = None

    if "fsed" in model_info["parameters"]:
        fsed = []
    else:
        fsed = None

    if "log_kzz" in model_info["parameters"]:
        log_kzz = []
    else:
        log_kzz = None

    if "ad_index" in model_info["parameters"]:
        ad_index = []
    else:
        ad_index = None

    flux = []

    print()

    if wavel_range is None:
        print(
            f"Wavelength range (um) = "
            f"{model_info['wavelength range'][0]} - "
            f"{model_info['wavelength range'][1]}"
        )

    else:
        print(f"Wavelength range (um) = {wavel_range[0]} - {wavel_range[1]}")

    if wavel_range is not None and wavel_sampling is not None:
        wavelength = create_wavelengths(wavel_range, wavel_sampling)
        resample_spectra = True

    else:
        wavelength = None
        resample_spectra = False
        wavel_sampling = model_info["lambda/d_lambda"]

    print(f"Sampling (lambda/d_lambda) = {wavel_sampling}")

    if teff_range is None:
        print(
            f"Teff range (K) = {model_info['teff range'][0]} - {model_info['teff range'][1]}"
        )
    else:
        print(f"Teff range (K) = {teff_range[0]} - {teff_range[1]}")

    print()
    print_message = ""

    for _, _, file_list in os.walk(data_folder):
        for file_name in sorted(file_list):
            if file_name[: len(model_tag)] == model_tag:
                file_split = file_name.split("_")

                param_index = file_split.index("teff") + 1
                teff_val = float(file_split[param_index])

                if teff_range is not None:
                    if teff_val < teff_range[0] or teff_val > teff_range[1]:
                        continue

                teff.append(teff_val)

                if logg is not None:
                    param_index = file_split.index("logg") + 1
                    logg.append(float(file_split[param_index]))

                if feh is not None:
                    param_index = file_split.index("feh") + 1
                    feh.append(float(file_split[param_index]))

                if c_o_ratio is not None:
                    param_index = file_split.index("co") + 1
                    c_o_ratio.append(float(file_split[param_index]))

                if fsed is not None:
                    param_index = file_split.index("fsed") + 1
                    fsed.append(float(file_split[param_index]))

                if log_kzz is not None:
                    param_index = file_split.index("logkzz") + 1
                    log_kzz.append(float(file_split[param_index]))

                if ad_index is not None:
                    param_index = file_split.index("adindex") + 1
                    ad_index.append(float(file_split[param_index]))

                empty_message = len(print_message) * " "
                print(f"\r{empty_message}", end="")

                print_message = (
                    f"Adding {model_info['name']} model spectra... {file_name}"
                )
                print(f"\r{print_message}", end="", flush=True)

                spec_file = os.path.join(data_folder, file_name)

                if file_name[-4:] == ".dat":
                    data_wavel, data_flux = np.loadtxt(spec_file, unpack=True)

                else:
                    data = np.load(spec_file)

                    data_wavel = data[:, 0]
                    data_flux = data[:, 1]

                if wavel_range is None:
                    if wavelength is None:
                        wavelength = np.copy(data_wavel)  # (um)

                        if np.all(np.diff(wavelength) < 0):
                            raise ValueError(
                                "The wavelengths are not all sorted by increasing value."
                            )

                    flux.append(data_flux)  # (W m-2 um-1)

                else:
                    if not resample_spectra:
                        if wavelength is None:
                            wavelength = np.copy(data_wavel)  # (um)

                            if np.all(np.diff(wavelength) < 0):
                                raise ValueError(
                                    "The wavelengths are not all sorted by increasing value."
                                )

                            wavel_select = (wavel_range[0] < wavelength) & (
                                wavelength < wavel_range[1]
                            )
                            wavelength = wavelength[wavel_select]

                        flux.append(data_flux[wavel_select])  # (W m-2 um-1)

                    else:
                        flux_resample = spectres.spectres(
                            wavelength,
                            data_wavel,
                            data_flux,
                            spec_errs=None,
                            fill=np.nan,
                            verbose=False,
                        )

                        if np.isnan(np.sum(flux_resample)):
                            raise ValueError(
                                f"Resampling is only possible if the new wavelength "
                                f"range ({wavelength[0]} - {wavelength[-1]} um) falls "
                                f"sufficiently far within the wavelength range "
                                f"({data_wavel[0]} - {data_wavel[-1]} um) of the input "
                                f"spectra."
                            )

                        flux.append(flux_resample)  # (W m-2 um-1)

    print()

    if logg is not None:
        logg = np.asarray(logg)

    if feh is not None:
        feh = np.asarray(feh)

    if c_o_ratio is not None:
        c_o_ratio = np.asarray(c_o_ratio)

    if fsed is not None:
        fsed = np.asarray(fsed)

    if log_kzz is not None:
        log_kzz = np.asarray(log_kzz)

    if ad_index is not None:
        ad_index = np.asarray(ad_index)

    data_sorted = sort_data(
        np.asarray(teff),
        logg,
        feh,
        c_o_ratio,
        fsed,
        log_kzz,
        ad_index,
        wavelength,
        np.asarray(flux),
    )

    write_data(
        model_tag, model_info["parameters"], wavel_sampling, database, data_sorted
    )

    add_missing(model_tag, model_info["parameters"], database)



"""
Utility functions for converting quantities.
"""

from typing import Optional, Tuple, Union

import numpy as np

from typeguard import typechecked

from species.core import constants


@typechecked
def apparent_to_absolute(
    app_mag: Union[
        Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]
    ],
    distance: Union[
        Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]
    ],
) -> Union[Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]]:
    """
    Function for converting an apparent magnitude into an absolute
    magnitude. The uncertainty on the distance is propagated into the
    uncertainty on the absolute magnitude.

    Parameters
    ----------
    app_mag : tuple(float, float), tuple(np.ndarray, np.ndarray)
        Apparent magnitude and uncertainty (mag). The returned error
        on the absolute magnitude is set to None if the error on the
        apparent magnitude is set to None, for example
        ``app_mag=(15., None)``.
    distance : tuple(float, float), tuple(np.ndarray, np.ndarray)
        Distance and uncertainty (pc). The error is not propagated
        into the error on the absolute magnitude if set to None, for
        example ``distance=(20., None)``.

    Returns
    -------
    float, np.ndarray
        Absolute magnitude (mag).
    float, np.ndarray, None
        Uncertainty (mag).
    """

    abs_mag = app_mag[0] - 5.0 * np.log10(distance[0]) + 5.0

    if app_mag[1] is not None and distance[1] is not None:
        dist_err = distance[1] * (5.0 / (distance[0] * np.log(10.0)))
        abs_err = np.sqrt(app_mag[1] ** 2 + dist_err**2)

    elif app_mag[1] is not None and distance[1] is None:
        abs_err = app_mag[1]

    else:
        abs_err = None

    return abs_mag, abs_err


@typechecked
def absolute_to_apparent(
    abs_mag: Union[
        Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]
    ],
    distance: Union[
        Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]
    ],
) -> Union[Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]]:
    """
    Function for converting an absolute magnitude
    into an apparent magnitude.

    Parameters
    ----------
    abs_mag : tuple(float, float), tuple(np.ndarray, np.ndarray)
        Tuple with the absolute magnitude and uncertainty (mag).
        The uncertainty on the returned apparent magnitude is
        simply adopted from the absolute magnitude. Providing the
        uncertainty is optional and can be set to ``None``.
    distance : tuple(float, float), tuple(np.ndarray, np.ndarray)
        Tuple with the distance and uncertainty (pc). The uncertainty
        is optional and can be set to ``None``. The distance
        uncertainty is currently not used by this function but
        included so it can be implemented at some point into the
        error budget.

    Returns
    -------
    float, np.ndarray
        Apparent magnitude (mag).
    float, np.ndarray, None
        Uncertainty (mag).
    """

    app_mag = abs_mag[0] + 5.0 * np.log10(distance[0]) - 5.0

    return app_mag, abs_mag[1]


@typechecked
def parallax_to_distance(
    parallax: Union[
        Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]
    ],
) -> Union[Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]]:
    """
    Function for converting from parallax to distance.

    Parameters
    ----------
    parallax : tuple(float, float), tuple(np.ndarray, np.ndarray)
        Parallax and optional uncertainty (mas). The
        uncertainty is not used if set to ``None``,
        for example, ``parallax=(2., None)``.

    Returns
    -------
    float, np.ndarray
        Distance (pc).
    float, np.ndarray, None
        Uncertainty (pc).
    """

    # From parallax (mas) to distance (pc)
    distance = 1e3 / parallax[0]

    if parallax[1] is None:
        distance_error = None

    else:
        distance_minus = distance - 1.0 / ((parallax[0] + parallax[1]) * 1e-3)
        distance_plus = 1.0 / ((parallax[0] - parallax[1]) * 1e-3) - distance
        distance_error = (distance_plus + distance_minus) / 2.0

    return distance, distance_error


@typechecked
def logg_to_mass(
    logg: Union[float, np.ndarray], radius: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Function for converting :math:`\\log(g)` and a radius into a mass.

    Parameters
    ----------
    logg : float, np.ndarray
        Log10 of the surface gravity (cgs).
    radius : float, np.ndarray
        Radius (Rjup).

    Returns
    -------
    float, np.ndarray
        Mass (Mjup).
    """

    surface_grav = 1e-2 * 10.0**logg  # (m s-2)
    radius *= constants.R_JUP  # (m)
    mass = surface_grav * radius**2 / constants.GRAVITY  # (kg)

    return mass / constants.M_JUP


@typechecked
def logg_to_radius(
    logg: Union[float, np.ndarray], mass: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Function for converting :math:`\\log(g)` and a mass into a radius.

    Parameters
    ----------
    logg : float, np.ndarray
        Log10 of the surface gravity (cgs).
    mass : float, np.ndarray
        Mass (Mjup).

    Returns
    -------
    float, np.ndarray
        Radius (Rjup).
    """

    surface_grav = 1e-2 * 10.0**logg  # (m s-2)
    mass_kg = mass * constants.M_JUP  # (kg)
    radius = np.sqrt(mass_kg * constants.GRAVITY / surface_grav)  # (m)

    return radius / constants.R_JUP


@typechecked
def mass_to_logg(
    mass: Union[float, np.ndarray], radius: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Function for converting a mass and radius into :math:`\\log(g)`.

    Parameters
    ----------
    mass : float, np.ndarray
        Mass ($M_\\mathrm{J}$).
    radius : float, np.ndarray
        Radius ($R_\\mathrm{J}$).
    Returns
    -------
    float, np.ndarray
        Surface gravity :math:`\\log(g)`.
    """

    mass *= constants.M_JUP  # (kg)
    radius *= constants.R_JUP  # (m)
    gravity = 1e2 * mass * constants.GRAVITY / radius**2  # (cm s-2)

    return np.log10(gravity)


@typechecked
def luminosity_to_teff(
    luminosity: Union[float, np.ndarray], radius: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Function for converting a luminosity and radius into :math:`T_\\mathrm{eff}`.
    Parameters
    ----------
    luminosity : float, np.ndarray
        Bolometric luminosity ($L_\\odot$).
    radius : float, np.ndarray
        Radius ($R_\\mathrm{J}$).
    Returns
    -------
    float, np.ndarray
        Effective temperature (K).
    """

    radius *= constants.R_JUP  # (Rjup)
    teff = (luminosity / (4.0 * np.pi * radius**2 * constants.SIGMA_SB)) ** 0.25

    return teff



"""
Module for setting up species in the working folder.
"""

import json
import os
import socket
import urllib.request

from configparser import ConfigParser
from importlib.util import find_spec
from typeguard import typechecked

import h5py
import species


class SpeciesInit:
    """
    Class for initiating species by creating the database and
    configuration file in case they are not present in the working
    folder, and creating the data folder for storage of input data.
    """

    @typechecked
    def __init__(self) -> None:
        """
        Returns
        -------
        NoneType
            None
        """

        species_version = species.__version__
        species_msg = f"species v{species_version}"

        print(len(species_msg) * "=")
        print(species_msg)
        print(len(species_msg) * "=")

        try:
            pypi_url = "https://pypi.org/pypi/species/json"

            with urllib.request.urlopen(pypi_url, timeout=1.0) as open_url:
                url_content = open_url.read()
                url_data = json.loads(url_content)
                latest_version = url_data["info"]["version"]

        except (urllib.error.URLError, socket.timeout):
            latest_version = None

        if latest_version is not None and species_version != latest_version:
            print(f"\n -> A new version ({latest_version}) is available!")
            print(" -> It is recommended to update to the latest version")
            print(" -> See https://github.com/tomasstolker/species for details")

        working_folder = os.path.abspath(os.getcwd())
        print(f"\nWorking folder: {working_folder}")

        config_file = os.path.join(working_folder, "species_config.ini")

        if os.path.isfile(config_file):
            print(f"\nConfiguration file: {config_file}")

        else:
            print("\nCreating species_config.ini...", end="", flush=True)

            with open(config_file, "w", encoding="utf-8") as file_obj:
                file_obj.write("[species]\n\n")

                file_obj.write("; File with the HDF5 database\n")
                file_obj.write("database = species_database.hdf5\n\n")

                file_obj.write("; Folder where data will be downloaded\n")
                file_obj.write("data_folder = ./data/\n\n")

                file_obj.write("; Magnitude of Vega for all filters\n")
                file_obj.write("vega_mag = 0.03\n")

            print(" [DONE]")

        config = ConfigParser()
        config.read(config_file)

        if "database" in config["species"]:
            database_file = os.path.abspath(config["species"]["database"])

        else:
            database_file = "species_database.hdf5"

            with open(config_file, "a", encoding="utf-8") as file_obj:
                file_obj.write("\n; File with the HDF5 database\n")
                file_obj.write("database = species_database.hdf5\n")

        if "data_folder" in config["species"]:
            data_folder = os.path.abspath(config["species"]["data_folder"])

        else:
            data_folder = "./data/"

            with open(config_file, "a", encoding="utf-8") as file_obj:
                file_obj.write("\n; Folder where data will be downloaded\n")
                file_obj.write("data_folder = ./data/\n")

        if "vega_mag" in config["species"]:
            vega_mag = config["species"]["vega_mag"]

        else:
            vega_mag = 0.03

            with open(config_file, "a", encoding="utf-8") as file_obj:
                file_obj.write("\n; Magnitude of Vega for all filters\n")
                file_obj.write("vega_mag = 0.03\n")

        if os.path.isfile(database_file):
            print(f"Database file: {database_file}")

        else:
            print("Creating species_database.hdf5...", end="", flush=True)
            h5_file = h5py.File(database_file, "w")
            h5_file.close()
            print(" [DONE]")

        if os.path.exists(data_folder):
            print(f"Data folder: {data_folder}")

        else:
            print("Creating data folder...", end="", flush=True)
            os.makedirs(data_folder)
            print(" [DONE]")

        print("\nConfiguration settings:")
        print(f"   - Database: {database_file}")
        print(f"   - Data folder: {data_folder}")
        print(f"   - Magnitude of Vega: {vega_mag}")

        if find_spec("mpi4py") is None:
            print("\nMultiprocessing: mpi4py not installed")

        else:
            from mpi4py import MPI

            # Rank of this process in a communicator
            mpi_rank = MPI.COMM_WORLD.Get_rank()

            # Number of processes in a communicator
            mpi_size = MPI.COMM_WORLD.Get_size()

            print("\nMultiprocessing: mpi4py installed")
            print(f"Process number {mpi_rank+1:d} out of {mpi_size:d}...")
